from udsoncan.Request import Request
from udsoncan.Response import Response
from udsoncan import MemoryLocation
from udsoncan.BaseService import BaseService, BaseResponseData
from udsoncan.ResponseCode import ResponseCode
from udsoncan.exceptions import *

from typing import cast


class WriteMemoryByAddress(BaseService):
    _sid = 0x3D
    _use_subfunction = False

    supported_negative_response = [ResponseCode.IncorrectMessageLengthOrInvalidFormat,
                                   ResponseCode.ConditionsNotCorrect,
                                   ResponseCode.RequestOutOfRange,
                                   ResponseCode.SecurityAccessDenied,
                                   ResponseCode.GeneralProgrammingFailure
                                   ]

    class ResponseData(BaseResponseData):
        """
        .. data:: alfid_echo

                AddressAndLengthFormatIdentifier used in the :ref:`MemoryLocation <MemoryLocation>` object echoed back by the server as a single byte integer.

        .. data:: memory_location_echo

                An instance of :ref:`MemoryLocation <MemoryLocation>` that includes the address, size and alfid that the server echoed back.		
        """

        alfid_echo: int
        memory_location_echo: MemoryLocation

        def __init__(self, alfid_echo: int, memory_location_echo: MemoryLocation):
            super().__init__(WriteMemoryByAddress)
            self.alfid_echo = alfid_echo
            self.memory_location_echo = memory_location_echo

    class InterpretedResponse(Response):
        service_data: "WriteMemoryByAddress.ResponseData"

    @classmethod
    def make_request(cls, memory_location: MemoryLocation, data: bytes) -> Request:
        """
        Generates a request for ReadMemoryByAddress

        :param memory_location: The address and the size of the memory block to write.
        :type memory_location: :ref:`MemoryLocation <MemoryLocation>`

        :param data: The data to write into memory.
        :type data: bytes

        :raises ValueError: If parameters are out of range, missing or wrong type
        """

        if not isinstance(memory_location, MemoryLocation):
            raise ValueError('Given memory location must be an instance of MemoryLocation')

        if not isinstance(data, bytes):
            raise ValueError('data must be a bytes object')
        request = Request(service=cls)

        request.data = bytes()
        request.data += memory_location.alfid.get_byte()  # AddressAndLengthFormatIdentifier
        request.data += memory_location.get_address_bytes()
        request.data += memory_location.get_memorysize_bytes()
        request.data += data

        return request

    @classmethod
    def interpret_response(cls, response: Response, memory_location: MemoryLocation) -> InterpretedResponse:
        """
        Populates the response ``service_data`` property with an instance of :class:`WriteMemoryByAddress.ResponseData<udsoncan.services.WriteMemoryByAddress.ResponseData>`

        :param response: The received response to interpret
        :type response: :ref:`Response<Response>`

        :param memory_location: The memory location used for the request. 
                The bytes position varies depending on the ``memory_location`` format
        :type memory_location: :ref:`MemoryLocation <MemoryLocation>`

        :raises InvalidResponseException: If length of ``response.data`` is too short
        """

        if not isinstance(memory_location, MemoryLocation):
            raise ValueError('Given memory location must be an instance of MemoryLocation')

        if response.data is None:
            raise InvalidResponseException(response, "No data in response")

        address_bytes = memory_location.get_address_bytes()
        memorysize_bytes = memory_location.get_memorysize_bytes()

        expected_response_size = 1 + len(address_bytes) + len(memorysize_bytes)
        if len(response.data) < expected_response_size:
            raise InvalidResponseException(response, 'Repsonse should be at least %d bytes' % (expected_response_size))

        offset = 1
        length = len(memory_location.get_address_bytes())
        address_echo = response.data[1:1 + length]
        offset += length
        length = len(memory_location.get_memorysize_bytes())
        memorysize_echo = response.data[offset:offset + length]

        response.service_data = cls.ResponseData(
            alfid_echo=response.data[0],
            memory_location_echo=MemoryLocation.from_bytes(address_bytes=address_echo, memorysize_bytes=memorysize_echo)
        )

        return cast(WriteMemoryByAddress.InterpretedResponse, response)



from udsoncan.exceptions import ConfigError
from udsoncan.BaseService import BaseService

from udsoncan.typing import IOConfigEntry, IOConfig, CodecDefinition

from typing import Any, Union, List, Dict, cast


def validate_int(value: Any, min: int = 0, max: int = 0xFF, name: str = 'value'):
    if not isinstance(value, int):
        raise ValueError("%s must be a valid integer" % (name))
    if value < min or value > max:
        raise ValueError("%s must be an integer between 0x%X and 0x%X" % (name, min, max))


def fetch_io_entry_from_config(did: int, ioconfig: IOConfig) -> IOConfigEntry:
    if did not in ioconfig:
        if 'default' in ioconfig:
            selected = ioconfig['default']
        else:
            raise ConfigError(key=did, msg='Actual Input/Output configuration contains no definition for data identifier 0x%04x' % did)
    else:
        selected = ioconfig[did]

    if not isinstance(selected, dict):
        selected = cast(IOConfigEntry, {
            'codec': selected
        })
    return selected


def check_io_config_composite_entry(didname: str, entry: IOConfigEntry) -> None:
    if 'codec'not in entry:
        raise ConfigError('codec', msg='Configuration for Input/Output identifier %s is missing a codec')

    if 'mask' in entry:
        mask_def = entry['mask']

        for mask_name in mask_def:
            if not isinstance(mask_def[mask_name], int):
                raise ValueError('In Input/Output configuration for did %s, mask "%s" is not an integer' % (didname, mask_name))

            if mask_def[mask_name] < 0:
                raise ValueError('In Input/Output configuration for did %s, mask "%s" is not a positive integer' % (didname, mask_name))

    if 'mask_size' in entry:
        if not isinstance(entry['mask_size'], int):
            raise ValueError('mask_size in Input/Output configuration for did %s must be a valid integer' % (didname))

        if entry['mask_size'] < 0:
            raise ValueError('mask_size in Input/Output configuration for did %s must be greater than 0' % (didname))

        if 'mask' in entry:
            mask_def = entry['mask']
            for mask_name in mask_def:
                if mask_def[mask_name] > 2**(entry['mask_size'] * 8) - 1:
                    raise ValueError(
                        'In Input/Output configuration for did %s, mask "%s" cannot fit in %d bytes (defined by mask_size)' % (didname, mask_name, entry['mask_size']))


# Make sure that the actual client configuration contains valid definitions for given Input/Output Data Identifiers
def check_io_config(didlist: Union[int, List[int]], ioconfig: Dict[Any, Any]) -> IOConfig:
    didlist = [didlist] if not isinstance(didlist, list) else didlist

    if 'input_output' in ioconfig:
        ioconfig = ioconfig['input_output']

    if not isinstance(ioconfig, dict):
        raise ConfigError('input_output', msg='Configuration of Input/Output section must be a dict.')

    for did in didlist:
        io_config_entry = fetch_io_entry_from_config(did, ioconfig)
        didstr = '"default"' if did not in ioconfig else '0x%04x' % did
        check_io_config_composite_entry(didstr, io_config_entry)

    return cast(IOConfig, ioconfig)



__all__ = [
    'IOMasks',
    'IOValues'
]

from typing import Dict, Any, Tuple


class IOMasks:
    """
    Allow to specify a list of masks for a :ref:`InputOutputControlByIdentifier<InputOutputControlByIdentifier>` composite codec.

    Example : IOMasks(mask1,mask2, mask3=True, mask4=False)

    :param args: Masks to set to True
    :param kwargs: Masks and their values
    """
    maskdict: Dict[str, bool]

    def __init__(self, *args: str, **kwargs: bool):
        for k in kwargs:
            if not isinstance(kwargs[k], bool):
                raise ValueError('mask value must be a boolean value')

        for k in args:
            if not isinstance(k, str):
                raise ValueError('Mask name must be a valid string')

        self.maskdict = dict()
        for k in args:
            self.maskdict[k] = True

        for k in kwargs:
            if not isinstance(kwargs[k], bool):
                raise ValueError('Mask value must be True or False')
            self.maskdict[k] = kwargs[k]

    def get_dict(self):
        return self.maskdict


# Used for IO Control service. Allows comprehensive one-liner.
class IOValues:
    """
    This class saves a function argument so they can be passed to a callback function.

    :param args: Arguments
    :param kwargs: Named arguments
    """

    args: Tuple
    kwargs: Dict[str, Any]

    def __init__(self, *args: Any, **kwargs: Any):
        self.args = args
        self.kwargs = kwargs



import struct
from udsoncan import DataFormatIdentifier, Filesize
from udsoncan.Request import Request
from udsoncan.Response import Response
from udsoncan.exceptions import *
from udsoncan.BaseService import BaseService, BaseSubfunction, BaseResponseData
from udsoncan.ResponseCode import ResponseCode

from typing import Optional, Union, cast


class RequestFileTransfer(BaseService):
    _sid = 0x38
    _use_subfunction = False

    supported_negative_response = [
        ResponseCode.IncorrectMessageLengthOrInvalidFormat,
        ResponseCode.ConditionsNotCorrect,
        ResponseCode.RequestOutOfRange,
        ResponseCode.UploadDownloadNotAccepted,
        ResponseCode.RequestSequenceError  # ResumeFile only
    ]

    class ModeOfOperation(BaseSubfunction):  # Not really a subfunction, but we wantto inherit the helpers in BaseSubfunction class
        """
        RequestFileTransfer Mode Of Operation (MOOP). Represent the action that can be done on the server filesystem.
        See ISO-14229:2013 Annex G
        """

        __pretty_name__ = 'mode of operation'

        AddFile = 1
        DeleteFile = 2
        ReplaceFile = 3
        ReadFile = 4
        ReadDir = 5
        ResumeFile = 6

    @classmethod
    def normalize_data_format_identifier(cls, dfi: Optional[DataFormatIdentifier]) -> DataFormatIdentifier:
        if dfi is None:
            dfi = DataFormatIdentifier()

        if not isinstance(dfi, DataFormatIdentifier):
            raise ValueError('dfi must be an instance of DataFormatIdentifier')

        return dfi

    class ResponseData(BaseResponseData):
        """
        .. data:: moop_echo (int)

                Request ModeOfOperation echoed back by the server

        .. data:: max_length (int)

                The MaxNumberOfBlockLength returned by the server. Represent the number of data bytes that should be included
                in each subsequent TransferData request excepted the last one that might be smaller. 

                Not set for a response to ``DeleteFile``.

        .. data:: dfi (DataFormatIdentifier)

                Request DataFormatIdentifier echoed back by the server.

                Not set for a response to ``DeleteFile``.
                Set to Compression=0, Encryption=0, when getting a response for ``ReadDir`` as specified by ISO-14229.

        .. data:: filesize (Filesize)

                Defines the size fo the file to be read in bytes, including its uncompressed and compressed size.

                Only set when performing a ``ReadFile`` request

        .. data:: dirinfo_length (int)

                Defines the size of the directory information to be read in bytes.

                Only set when performing a ``ReadDir`` request

        .. data:: fileposition (int)

                Defines the position of the at which the tester will resume downloading after an initial download is suspended.

                Only set when performing a ``ResumeFile`` request

        """
        __slots__ = 'moop_echo', 'max_length', 'dfi', 'filesize', 'dirinfo_length', 'fileposition'

        moop_echo: int
        max_length: Optional[int]
        dfi: Optional[DataFormatIdentifier]
        filesize: Optional[Filesize]
        dirinfo_length: Optional[int]
        fileposition: Optional[int]

        def __init__(self,
                     moop_echo: int,
                     max_length: Optional[int] = None,
                     dfi: Optional[DataFormatIdentifier] = None,
                     filesize: Optional[Filesize] = None,
                     dirinfo_length: Optional[int] = None,
                     fileposition: Optional[int] = None):
            super().__init__(RequestFileTransfer)
            self.moop_echo = moop_echo
            self.max_length = max_length
            self.dfi = dfi
            self.filesize = filesize
            self.dirinfo_length = dirinfo_length
            self.fileposition = fileposition

    class InterpretedResponse(Response):
        service_data: "RequestFileTransfer.ResponseData"

    @classmethod
    def make_request(cls,
                     moop: int,
                     path: str,
                     dfi: Optional[DataFormatIdentifier] = None,
                     filesize: Optional[Union[Filesize, int]] = None
                     ) -> Request:
        """
        Generates a request for RequestFileTransfer

        :param moop: Mode of operation. Can be AddFile(1), DeleteFile(2), ReplaceFile(3), ReadFile(4), ReadDir(5), ResumeFile(6). See :class:`RequestFileTransfer.ModeOfOperation<udsoncan.services.RequestFileTransfer.ModeOfOperation>`
        :type moop: int

        :param path: String representing the path to the target file or directory.
        :type path: string

        :param dfi: DataFormatIdentifier defining the compression and encryption scheme of the data.
                If not specified, the default value of 00 will be used, specifying no encryption and no compression.
                This value is only used when ModeOfOperation is : ``AddFile``, ``ReplaceFile``, ``ReadFile``, ``ResumeFile``
        :type dfi: :ref:`DataFormatIdentifier<DataFormatIdentifier>`

        :param filesize: The filesize of the file to write when ModeOfOperation is ``AddFile``, ``ReplaceFile`` or ``ResumeFile``.
            If filesize is an object of type :ref:`Filesize<Filesize>`, the uncompressed size and compressed size will be encoded on
            the minimum amount of bytes necessary, unless a ``width`` is explicitly defined. If no compressed size is given or filesize is an ``int``,
            then the compressed size will be set equal to the uncompressed size or the integer value given as specified by ISO-14229
        :type filesize: :ref:`Filesize<Filesize>` or int

        :raises ValueError: If parameters are out of range, missing or wrong type
        """
        if not isinstance(moop, int):
            raise ValueError('Mode of operation must be an integer')

        if moop not in [cls.ModeOfOperation.AddFile,
                        cls.ModeOfOperation.DeleteFile,
                        cls.ModeOfOperation.ReplaceFile,
                        cls.ModeOfOperation.ReadFile,
                        cls.ModeOfOperation.ReadDir,
                        cls.ModeOfOperation.ResumeFile]:
            raise ValueError("Mode of operation of %d is not a known mode" % moop)

        if not isinstance(path, str):
            raise ValueError('Given path must be a valid string')

        if len(path) <= 0:
            raise ValueError('Path must be a string longer than 0 character')

        path_ascii = path.encode('ascii')
        if len(path_ascii) > 0xFFFF:
            raise ValueError('Path length must be smaller or equal than 65535  bytes (16 bits) when encoded in ASCII')

        use_dfi = moop in [cls.ModeOfOperation.AddFile, cls.ModeOfOperation.ReplaceFile, cls.ModeOfOperation.ReadFile, cls.ModeOfOperation.ResumeFile]
        use_filesize = moop in [cls.ModeOfOperation.AddFile, cls.ModeOfOperation.ReplaceFile, cls.ModeOfOperation.ResumeFile]

        if use_dfi:
            dfi = cls.normalize_data_format_identifier(dfi)
        else:
            if dfi is not None:
                raise ValueError('DataFormatIdentifier is not needed with ModeOfOperation=%d' % moop)

        if use_filesize:
            if filesize is None:
                raise ValueError('A filesize must be given for this mode of operation')

            if isinstance(filesize, int):
                filesize = Filesize(filesize)

            if not isinstance(filesize, Filesize):
                raise ValueError('Given filesize must be a valid Filesize object or an integer')

            if filesize.uncompressed is None:
                raise ValueError('Filesize needs at least an Uncompressed file size')

            if filesize.compressed is None:
                filesize = Filesize(uncompressed=filesize.uncompressed, compressed=filesize.uncompressed, width=filesize.get_width())
        else:
            if filesize is not None:
                raise ValueError('Filesize is not needed with ModeOfOperation=%d' % moop)

        data = moop.to_bytes(1, 'big')
        data += len(path_ascii).to_bytes(2, 'big')
        data += path_ascii
        if use_dfi:
            assert dfi is not None  # mypy nitpick
            data += dfi.get_byte()
        if use_filesize:
            assert filesize is not None  # mypy nitpick
            data += filesize.get_width().to_bytes(1, 'big')
            data += filesize.get_uncompressed_bytes()
            data += filesize.get_compressed_bytes()

        request = Request(cls, data=data)
        return request

    @classmethod
    def interpret_response(cls, response: Response, tolerate_zero_padding: bool = True) -> InterpretedResponse:
        """
        Populates the response ``service_data`` property with an instance of :class:`RequestFileTransfer.ResponseData<udsoncan.services.RequestFileTransfer.ResponseData>`

        :param response: The received response to interpret
        :type response: :ref:`Response<Response>`

        :raises InvalidResponseException: If length of ``response.data`` is too short or payload does not respect ISO-14229 specifications
        :raises NotImplementedError: If the MaxNumberOfBlock or fileSizeUncompressedOrDirInfoLength value is encoded over more than 8 bytes.
        """
        if response.data is None:
            raise InvalidResponseException(response, "No data in response")

        if len(response.data) < 1:
            raise InvalidResponseException(response, 'Response payload must be at least 1 byte long')

        response.service_data = cls.ResponseData(
            moop_echo=int(response.data[0])
        )

        has_lfid = response.service_data.moop_echo in [cls.ModeOfOperation.AddFile, cls.ModeOfOperation.ResumeFile,
                                                       cls.ModeOfOperation.ReplaceFile, cls.ModeOfOperation.ReadFile, cls.ModeOfOperation.ReadDir]
        has_dfi = response.service_data.moop_echo in [cls.ModeOfOperation.AddFile, cls.ModeOfOperation.ResumeFile,
                                                      cls.ModeOfOperation.ReplaceFile, cls.ModeOfOperation.ReadFile, cls.ModeOfOperation.ReadDir]
        has_filesize_length = response.service_data.moop_echo in [cls.ModeOfOperation.ReadFile, cls.ModeOfOperation.ReadDir]
        has_uncompressed_filesize = response.service_data.moop_echo in [cls.ModeOfOperation.ReadFile, cls.ModeOfOperation.ReadDir]
        has_compressed_filesize = response.service_data.moop_echo in [cls.ModeOfOperation.ReadFile]
        has_fileposition = response.service_data.moop_echo in [cls.ModeOfOperation.ResumeFile]

        cursor = 1
        if has_lfid:
            if len(response.data) < 2:
                raise InvalidResponseException(
                    response, 'Response payload must be at least 2 byte long for Mode of operation %d' % response.service_data.moop_echo)
            lfid = int(response.data[1])
            cursor = 2

            if lfid > 8:
                raise NotImplementedError(
                    'This client does not support number bigger than %d bits, but MaxNumberOfBlock is encoded on %d bits' % ((8 * 8), (lfid * 8)))

            if lfid == 0:
                raise InvalidResponseException(response, 'Received a MaxNumberOfBlockLength of 0 which is impossible')

            if len(response.data) < 2 + lfid:
                raise InvalidResponseException(
                    response, 'Response payload says that MaxNumberOfBlock is encoded on %d bytes, but only %d bytes are present' % (lfid, (len(response.data) - 2)))

            todecode = bytearray(b'\x00\x00\x00\x00\x00\x00\x00\x00')
            for i in range(1, lfid + 1):
                todecode[-i] = response.data[cursor + lfid - i]
            response.service_data.max_length = struct.unpack('>q', todecode)[0]
            cursor += lfid

        if has_dfi:
            if len(response.data) < cursor + 1:
                raise InvalidResponseException(response, 'Missing DataFormatIdentifier in received response')

            dfi = DataFormatIdentifier.from_byte(response.data[cursor])
            response.service_data.dfi = dfi
            cursor += 1
            dfi_int = dfi.get_byte_as_int()

            if response.service_data.moop_echo == cls.ModeOfOperation.ReadDir and dfi_int != 0:
                raise InvalidResponseException(
                    response, 'DataFormatIdentifier for ReadDir can only be 0x00 as per ISO-14229, but its value was set to 0x%02x' % (dfi_int))

        if has_filesize_length:
            if len(response.data) < cursor + 2:
                raise InvalidResponseException(response, 'Missing or incomplete FileSizeOrDirInfoParameterLength in received response')
            fsodipl = struct.unpack('>H', response.data[cursor:cursor + 2])[0]
            cursor += 2

            if fsodipl > 8:
                raise NotImplementedError(
                    response, 'This client does not support number bigger than %d bits, but FileSizeOrDirInfoLength is encoded on %d bits' % ((8 * 8), (fsodipl * 8)))

            if fsodipl == 0:
                raise InvalidResponseException(response, 'Received a FileSizeOrDirInfoParameterLength of 0 which is impossible')

            if has_uncompressed_filesize:
                if len(response.data) < cursor + fsodipl:
                    raise InvalidResponseException(response, 'Missing or incomplete fileSizeUncompressedOrDirInfoLength in received response')

                todecode = bytearray(b'\x00\x00\x00\x00\x00\x00\x00\x00')
                for i in range(1, fsodipl + 1):
                    todecode[-i] = response.data[cursor + fsodipl - i]
                uncompressed_size = struct.unpack('>q', todecode)[0]
                cursor += fsodipl
            else:
                uncompressed_size = None

            if has_compressed_filesize:
                if len(response.data) < cursor + fsodipl:
                    raise InvalidResponseException(response, 'Missing or incomplete fileSizeCompressed in received response')

                todecode = bytearray(b'\x00\x00\x00\x00\x00\x00\x00\x00')
                for i in range(1, fsodipl + 1):
                    todecode[-i] = response.data[cursor + fsodipl - i]
                compressed_size = struct.unpack('>q', todecode)[0]
                cursor += fsodipl
            else:
                compressed_size = None

        if has_uncompressed_filesize and response.service_data.moop_echo == cls.ModeOfOperation.ReadDir:
            response.service_data.dirinfo_length = uncompressed_size
        else:
            if has_uncompressed_filesize or has_compressed_filesize:
                response.service_data.filesize = Filesize(uncompressed=uncompressed_size, compressed=compressed_size)

        if has_fileposition:
            fposl = 8  # standard has hardcoded number of bytes to 8
            if len(response.data) < cursor + fposl:
                raise InvalidResponseException(response, 'Missing or incomplete FilePosition in received response')

            todecode = bytearray(b'\x00\x00\x00\x00\x00\x00\x00\x00')
            for i in range(1, fposl + 1):
                todecode[-i] = response.data[cursor + fposl - i]

            response.service_data.fileposition = struct.unpack('>q', todecode)[0]
            cursor += fposl

        if len(response.data) > cursor:
            if response.data[cursor:] == b'\x00' * (len(response.data) - cursor) and tolerate_zero_padding:
                pass
            else:
                raise InvalidResponseException(response, 'Response payload has extra data that has no meaning')

        return cast(RequestFileTransfer.InterpretedResponse, response)



import ctypes
from ctypes import Structure, WINFUNCTYPE, POINTER, cast, c_long, c_void_p, c_ulong, byref  # type: ignore

from enum import Enum

import logging


class PASSTHRU_MSG(Structure):
    _fields_ = [("ProtocolID", c_ulong),
                ("RxStatus", c_ulong),
                ("TxFlags", c_ulong),
                ("Timestamp", c_ulong),
                ("DataSize", c_ulong),
                ("ExtraDataIndex", c_ulong),
                ("Data", ctypes.c_ubyte * 4128)]


class SCONFIG(Structure):
    _fields_ = [("Parameter", c_ulong),
                ("Value", c_ulong)]


class SCONFIG_LIST(Structure):
    _fields_ = [("NumOfParams", c_ulong),
                ("ConfigPtr", POINTER(SCONFIG))]

    def __init__(self, values):
        self.NumOfParams = len(values)
        self.ConfigPtr = (SCONFIG * self.NumOfParams)(*values)


class J2534():
    dllPassThruOpen = None
    dllPassThruClose = None
    dllPassThruConnect = None
    dllPassThruDisconnect = None
    dllPassThruReadMsgs = None
    dllPassThruWriteMsgs = None
    dllPassThruStartPeriodicMsg = None
    dllPassThruStopPeriodicMsg = None
    dllPassThruReadVersion = None
    dllPassThruGetLastError = None
    dllPassThruStartMsgFilter = None
    dllPassThruIoctl = None

    def __init__(self, windll, rxid, txid):
        global dllPassThruOpen
        global dllPassThruClose
        global dllPassThruConnect
        global dllPassThruDisconnect
        global dllPassThruReadMsgs
        global dllPassThruWriteMsgs
        global dllPassThruStartPeriodicMsg
        global dllPassThruStopPeriodicMsg
        global dllPassThruReadVersion
        global dllPassThruGetLastError
        global dllPassThruStartMsgFilter
        global dllPassThruIoctl

        self.hDLL = ctypes.cdll.LoadLibrary(windll)
        self.rxid = rxid.to_bytes(4, 'big')
        self.txid = txid.to_bytes(4, 'big')
         # Determine mode ID29 or ID11
        self.txConnectFlags = TxStatusFlag.ISO15765_CAN_ID_29.value if txid >> 11 else TxStatusFlag.ISO15765_CAN_ID_11.value
        self.txFlags = self.txConnectFlags | TxStatusFlag.ISO15765_FRAME_PAD.value

        self.logger = logging.getLogger()

        dllPassThruOpenProto = WINFUNCTYPE(
            c_long,
            c_void_p,
            POINTER(c_ulong))

        dllPassThruOpenParams = (1, "pName", 0), (1, "pDeviceID", 0)
        dllPassThruOpen = dllPassThruOpenProto(("PassThruOpen", self.hDLL), dllPassThruOpenParams)

        dllPassThruCloseProto = WINFUNCTYPE(
            c_long,
            c_ulong)

        dllPassThruCloseParams = (1, "DeviceID", 0),
        dllPassThruClose = dllPassThruCloseProto(("PassThruClose", self.hDLL), dllPassThruCloseParams)

        dllPassThruConnectProto = WINFUNCTYPE(
            c_long,
            c_ulong,
            c_ulong,
            c_ulong,
            c_ulong,
            POINTER(c_ulong))

        dllPassThruConnectParams = (1, "DeviceID", 0), (1, "ProtocolID", 0), (1, "Flags", 0), (1, "BaudRate", 500000), (1, "pChannelID", 0)
        dllPassThruConnect = dllPassThruConnectProto(("PassThruConnect", self.hDLL), dllPassThruConnectParams)

        dllPassThruDisconnectProto = WINFUNCTYPE(
            c_long,
            c_ulong)

        dllPassThruDisconnectParams = (1, "ChannelID", 0),
        dllPassThruDisconnect = dllPassThruDisconnectProto(("PassThruDisconnect", self.hDLL), dllPassThruDisconnectParams)

        dllPassThruReadMsgsProto = WINFUNCTYPE(
            c_long,
            c_ulong,
            POINTER(PASSTHRU_MSG),
            POINTER(c_ulong),
            c_ulong)

        dllPassThruReadMsgsParams = (1, "ChannelID", 0), (1, "pMsg", 0), (1, "pNumMsgs", 0), (1, "Timeout", 0)
        dllPassThruReadMsgs = dllPassThruReadMsgsProto(("PassThruReadMsgs", self.hDLL), dllPassThruReadMsgsParams)

        dllPassThruWriteMsgsProto = WINFUNCTYPE(
            c_long,
            c_ulong,
            POINTER(PASSTHRU_MSG),
            POINTER(c_ulong),
            c_ulong)

        dllPassThruWriteMsgsParams = (1, "ChannelID", 0), (1, "pMsg", 0), (1, "pNumMsgs", 0), (1, "Timeout", 0)
        dllPassThruWriteMsgs = dllPassThruWriteMsgsProto(("PassThruWriteMsgs", self.hDLL), dllPassThruWriteMsgsParams)

        dllPassThruStartPeriodicMsgProto = WINFUNCTYPE(
            c_long,
            c_ulong,
            POINTER(PASSTHRU_MSG),
            POINTER(c_ulong),
            c_ulong)

        dllPassThruStartPeriodicMsgParams = (1, "ChannelID", 0), (1, "pMsg", 0), (1, "pMsgID", 0), (1, "TimeInterval", 0)
        dllPassThruStartPeriodicMsg = dllPassThruStartPeriodicMsgProto(("PassThruStartPeriodicMsg", self.hDLL), dllPassThruStartPeriodicMsgParams)

        dllPassThruStopPeriodicMsgProto = WINFUNCTYPE(
            c_long,
            c_ulong,
            c_ulong)

        dllPassThruStopPeriodicMsgParams = (1, "ChannelID", 0), (1, "MsgID", 0)
        dllPassThruStopPeriodicMsg = dllPassThruStopPeriodicMsgProto(("PassThruStopPeriodicMsg", self.hDLL), dllPassThruStopPeriodicMsgParams)

        dllPassThruReadVersionProto = WINFUNCTYPE(
            c_long,
            c_ulong,
            POINTER(ctypes.c_char),
            POINTER(ctypes.c_char),
            POINTER(ctypes.c_char))

        dllPassThruReadVersionParams = (1, "DeviceID", 0), (1, "pFirmwareVersion", 0), (1, "pDllVersion", 0), (1, "pApiVersoin", 0)
        dllPassThruReadVersion = dllPassThruReadVersionProto(("PassThruReadVersion", self.hDLL), dllPassThruReadVersionParams)

        dllPassThruGetLastErrorProto = WINFUNCTYPE(
            c_long,
            POINTER(ctypes.c_char),
        )
        dllPassThruGetLastErrorParams = (1, "pErrorDescription", 0),
        dllPassThruGetLastError = dllPassThruGetLastErrorProto(("PassThruGetLastError", self.hDLL), dllPassThruGetLastErrorParams)

        dllPassThruStartMsgFilterProto = WINFUNCTYPE(
            c_long,
            c_ulong,
            c_ulong,
            POINTER(PASSTHRU_MSG),
            POINTER(PASSTHRU_MSG),
            POINTER(PASSTHRU_MSG),
            POINTER(c_ulong)
        )

        dllPassThruStartMsgFilterParams = (1,"ChannelID",0), (1,"FilterType",0),(1,"pMaskMsg",0),(1,"pPatternMsg",0),(1,"pFlowControlMsg",0),(1,"pMsgID",0)

        dllPassThruStartMsgFilter = dllPassThruStartMsgFilterProto(("PassThruStartMsgFilter", self.hDLL), dllPassThruStartMsgFilterParams)

        dllPassThruIoctlProto = WINFUNCTYPE(
            c_long,
            c_ulong,
            c_ulong,
            c_void_p,
            c_void_p
        )

        dllPassThruIoctlParams = (1, "Handle", 0), (1, "IoctlID", 0), (1, "pInput", 0), (1, "pOutput", 0)

        dllPassThruIoctl = dllPassThruIoctlProto(("PassThruIoctl", self.hDLL), dllPassThruIoctlParams)

    def PassThruOpen(self, pDeviceID=None):
        if not pDeviceID:
            pDeviceID = ctypes.c_ulong()

        result = dllPassThruOpen(bytes('J2534-2:', 'ascii'), byref(pDeviceID))
        return Error_ID(hex(result)), pDeviceID

    def PassThruConnect(self, deviceID, protocol, baudrate, pChannelID=None):
        if not pChannelID:
            pChannelID = c_ulong()

        result = dllPassThruConnect(deviceID, protocol, self.txConnectFlags, baudrate, byref(pChannelID))
        return Error_ID(hex(result)), pChannelID

    def PassThruClose(self, DeviceID):
        result = dllPassThruClose(DeviceID)
        return Error_ID(hex(result))

    def PassThruDisconnect(self, ChannelID):
        result = dllPassThruDisconnect(ChannelID)
        return Error_ID(hex(result))

    def PassThruReadMsgs(self, ChannelID, protocol, pNumMsgs=1, Timeout=100):
        pMsg = PASSTHRU_MSG()
        pMsg.ProtocolID = protocol

        pNumMsgs = c_ulong(pNumMsgs)

        while 1:
            # breakpoint()
            result = dllPassThruReadMsgs(ChannelID, byref(pMsg), byref(pNumMsgs), c_ulong(Timeout))
            if Error_ID(hex(result)) == Error_ID.ERR_BUFFER_EMPTY or pNumMsgs == 0:
                return None, None, 0
            elif pMsg.RxStatus == 0 or pMsg.RxStatus == 0x100:
                return Error_ID(hex(result)), bytes(pMsg.Data[4:pMsg.DataSize]), pNumMsgs

    def PassThruWriteMsgs(self, ChannelID, Data, protocol, pNumMsgs=1, Timeout=1000):
        txmsg = PASSTHRU_MSG()
        txmsg.TxFlags = self.txFlags
        txmsg.ProtocolID = protocol;

        Data = self.txid + Data
        self.logger.info("Sending data: " + str(Data.hex()))

        for i in range(0, len(Data)):
            txmsg.Data[i] = Data[i]

        txmsg.DataSize = len(Data)

        result = dllPassThruWriteMsgs(ChannelID, byref(txmsg), byref(c_ulong(pNumMsgs)), c_ulong(Timeout))

        return Error_ID(hex(result))

    def PassThruStartPeriodicMsg(self, ChannelID, Data, MsgID=0, TimeInterval=100):
        pMsg = PASSTHRU_MSG()

        pMsg.Data = Data
        pMsg.DataSize = len(Data)

        result = dllPassThruStartPeriodicMsg(ChannelID, byref(pMsg), byref(c_ulong(MsgID)), c_ulong(TimeInterval))

        return Error_ID(hex(result))

    def PassThruStopPeriodicMsg(self, ChannelID, MsgID):
        result = dllPassThruStopPeriodicMsg(ChannelID, MsgID)

        return Error_ID(hex(result))

    def PassThruReadVersion(self, DeviceID):
        pFirmwareVersion = (ctypes.c_char * 80)()
        pDllVersion = (ctypes.c_char * 80)()
        pApiVersion = (ctypes.c_char * 80)()
        result = dllPassThruReadVersion(DeviceID, pFirmwareVersion, pDllVersion, pApiVersion)

        return Error_ID(hex(result)), pFirmwareVersion, pDllVersion, pApiVersion

    def PassThruGetLastError(self):
        pErrorDescription = (ctypes.c_char * 80)()
        result = dllPassThruGetLastError(pErrorDescription)

        return Error_ID(hex(result)), pErrorDescription.value.decode()

    def PassThruIoctl(self, Handle, IoctlID, ioctlInput=None, ioctlOutput=None):
        pInput = None if ioctlInput is None else byref(ioctlInput)
        pOutput = None if ioctlOutput is None else byref(ioctlOutput)

        result = dllPassThruIoctl(Handle, c_ulong(IoctlID.value), pInput, pOutput)

        return Error_ID(hex(result))

    def PassThruStartMsgFilter(self, ChannelID, protocol):
        msgMask = PASSTHRU_MSG()
        msgMask.ProtocolID = protocol
        msgMask.TxFlags = self.txFlags
        msgMask.DataSize = 4
        msgMask.RxStatus = msgMask.ExtraDataIndex = 0xCCCC_CCCC
        for i in range(0, 4):
            msgMask.Data[i] = 0xFF

        msgPattern = PASSTHRU_MSG()
        msgPattern.ProtocolID = protocol
        msgPattern.TxFlags = self.txFlags
        msgPattern.DataSize = 4
        msgPattern.RxStatus = msgPattern.ExtraDataIndex = 0xCCCC_CCCC
        for i in range(0, len(self.rxid)):
            msgPattern.Data[i] = self.rxid[i]

        msgFlow = PASSTHRU_MSG()
        msgFlow.ProtocolID = protocol;
        msgFlow.TxFlags = self.txFlags
        msgFlow.DataSize = 4
        msgFlow.RxStatus = msgFlow.ExtraDataIndex = 0xCCCC_CCCC
        for i in range(0, len(self.txid)):
            msgFlow.Data[i] = self.txid[i]

        filterType = c_ulong(Filter.FLOW_CONTROL_FILTER.value)
        msgID = c_ulong(0)

        result = dllPassThruStartMsgFilter(ChannelID, filterType, byref(msgMask), byref(msgPattern), byref(msgFlow), byref(msgID))

        return Error_ID(hex(result))


class Error_ID(Enum):
    ERR_SUCCESS = hex(0x00)
    STATUS_NOERROR = hex(0x00)
    ERR_NOT_SUPPORTED = hex(0x01)
    ERR_INVALID_CHANNEL_ID = hex(0x02)
    ERR_INVALID_PROTOCOL_ID = hex(0x03)
    ERR_NULL_PARAMETER = hex(0x04)
    ERR_INVALID_IOCTL_VALUE = hex(0x05)
    ERR_INVALID_FLAGS = hex(0x06)
    ERR_FAILED = hex(0x07)
    ERR_DEVICE_NOT_CONNECTED = hex(0x08)
    ERR_TIMEOUT = hex(0x09)
    ERR_INVALID_MSG = hex(0x0A)
    ERR_INVALID_TIME_INTERVAL = hex(0x0B)
    ERR_EXCEEDED_LIMIT = hex(0x0C)
    ERR_INVALID_MSG_ID = hex(0x0D)
    ERR_DEVICE_IN_USE = hex(0x0E)
    ERR_INVALID_IOCTL_ID = hex(0x0F)
    ERR_BUFFER_EMPTY = hex(0x10)
    ERR_BUFFER_FULL = hex(0x11)
    ERR_BUFFER_OVERFLOW = hex(0x12)
    ERR_PIN_INVALID = hex(0x13)
    ERR_CHANNEL_IN_USE = hex(0x14)
    ERR_MSG_PROTOCOL_ID = hex(0x15)
    ERR_INVALID_FILTER_ID = hex(0x16)
    ERR_NO_FLOW_CONTROL = hex(0x17)
    ERR_NOT_UNIQUE = hex(0x18)
    ERR_INVALID_BAUDRATE = hex(0x19)
    ERR_INVALID_DEVICE_ID = hex(0x1A)


class Protocol_ID(Enum):
    J1850VPW = 1
    J1850PWM = 2
    ISO9141 = 3
    ISO14230 = 4
    CAN = 5
    ISO15765 = 6
    SCI_A_ENGINE = 7  # OP2.0: Not supported
    SCI_A_TRANS = 8  # OP2.0: Not supported
    SCI_B_ENGINE = 9  # OP2.0: Not supported
    SCI_B_TRANS = 10  # OP2.0: Not supported


class Filter(Enum):
    PASS_FILTER = 0x00000001
    BLOCK_FILTER = 0x00000002
    FLOW_CONTROL_FILTER = 0x00000003


class TxStatusFlag(Enum):
    ISO15765_CAN_ID_BOTH = 0x00000800
    ISO15765_CAN_ID_29 = 0x00000100
    ISO15765_CAN_ID_11 = 0x00000040
    ISO15765_FRAME_PAD = 0x00000040
    WAIT_P3_MIN_ONLY = 0x00000200
    SW_CAN_HV_TX = 0x00000400  # OP2.0: Not supported
    SCI_MODE = 0x00400000  # OP2.0: Not supported
    SCI_TX_VOLTAGE = 0x00800000  # OP2.0: Not supported


class Ioctl_ID(Enum):
    GET_CONFIG = 0x01
    SET_CONFIG = 0x02
    READ_VBATT = 0x03
    FIVE_BAUD_INIT = 0x04
    FAST_INIT = 0x05
    CLEAR_TX_BUFFER = 0x07
    CLEAR_RX_BUFFER = 0x08
    CLEAR_PERIODIC_MSGS = 0x09
    CLEAR_MSG_FILTERS = 0x0A
    CLEAR_FUNCT_MSG_LOOKUP_TABLE = 0x0B
    ADD_TO_FUNCT_MSG_LOOKUP_TABLE = 0x0C
    DELETE_FROM_FUNCT_MSG_LOOKUP_TABLE = 0x0D
    READ_PROG_VOLTAGE = 0x0E

    DATA_RATE = 0x01  # 5 500000 	# Baud rate value used for vehicle network. No default value specified.
    LOOPBACK = 0x03  # 0(OFF)/1(ON)	# 0 = Do not echo transmitted messages to the Receive queue. 1 = Echo transmitted messages to the Receive queue.
    NODE_ADDRESS = 0x04  # 0x00-0xFF	# J1850PWM specific, physical address for node of interest in the vehicle network. Default is no nodes are recognized by scan tool.
    NETWORK_LINE = 0x05  # 0(BUS_NORMAL)/1(BUS_PLUS)/2(BUS_MINUS)	# J1850PWM specific, network line(s) active during message transfers. Default value is 0(BUS_NORMAL).
    P1_MIN = 0x06  # 0x0-0xFFFF	# ISO-9141/14230 specific, min. ECU inter-byte time for responses [02.02-API: ms]. Default value is 0 ms. 04.04-API: NOT ADJUSTABLE, 0ms.
    P1_MAX = 0x07  # 0x0/0x1-0xFFFF # ISO-9141/14230 specific, max. ECU inter-byte time for responses [02.02-API: ms, 04.04-API: *0.5ms]. Default value is 20 ms.
    P2_MIN = 0x08  # 0x0-0xFFFF	# ISO-9141/14230 specific, min. ECU response time to a tester request or between ECU responses [02.02-API: ms, 04.04-API: *0.5ms]. 04.04-API: NOT ADJUSTABLE, 0ms. Default value is 25 ms.
    P2_MAX = 0x09  # 0x0-0xFFFF	# ISO-9141/14230 specific, max. ECU response time to a tester request or between ECU responses [02.02-API: ms, 04.04-API: *0.5ms]. 04.04-API: NOT ADJUSTABLE, all messages up to P3_MIN are receoved. Default value is 50 ms.
    P3_MIN = 0x0A  # 0x0-0xFFFF	# ISO-9141/14230 specific, min. ECU response time between end of ECU response and next tester request [02.02-API: ms, 04.04-API: *0.5ms]. Default value is 55 ms.
    P3_MAX = 0x0B  # 0x0-0xFFFF	# ISO-9141/14230 specific, max. ECU response time between end of ECU response and next tester request [02.02-API: ms, 04.04-API: *0.5ms]. 04.04-API: NOT ADJUSTABLE, messages can be sent at anytime after P3_MIN. Default value is 5000 ms.
    P4_MIN = 0x0C  # 0x0-0xFFFF	# ISO-9141/14230 specific, min. tester inter-byte time for a request [02.02-API: ms, 04.04-API: *0.5ms]. Default value is 5 ms.
    P4_MAX = 0x0D  # 0x0-0xFFFF	# ISO-9141/14230 specific, max. tester inter-byte time for a request [02.02-API: ms, 04.04-API: *0.5ms]. 04.04-API: NOT ADJUSTABLE, P4_MIN is always used. Default value is 20 ms.
    W1 = 0x0E  # 0x0-0xFFFF	# ISO 9141 specific, max. time [ms] from the address byte end to synchronization pattern start. Default value is 300 ms.
    W2 = 0x0F  # 0x0-0xFFFF	# ISO 9141 specific, max. time [ms] from the synchronization byte end to key byte 1 start. Default value is 20 ms.
    W3 = 0x10  # 0x0-0xFFFF	# ISO 9141 specific, max. time [ms] between key byte 1 and key byte 2. Default value is 20 ms.
    W4 = 0x11  # 0x0-0xFFFF	# ISO 9141 specific, 02.02-API: max. time [ms] between key byte 2 and its inversion from the tester. Default value is 50 ms.
    W5 = 0x12  # 0x0-0xFFFF	# ISO 9141 specific, min. time [ms] before the tester begins retransmission of the address byte. Default value is 300 ms.
    TIDLE = 0x13  # 0x0-0xFFFF	# ISO 9141 specific, bus idle time required before starting a fast initialization sequence. Default value is W5 value.
    TINL = 0x14  # 0x0-0xFFFF	# ISO 9141 specific, the duration [ms] of the fast initialization low pulse. Default value is 25 ms.
    TWUP = 0x15  # 0x0-0xFFFF	# ISO 9141 specific, the duration [ms] of the fast initialization wake-up pulse. Default value is 50 ms.
    PARITY = 0x16  # 0(NO_PARITY)/1(ODD_PARITY)/2(EVEN_PARITY)	# ISO9141 specific, parity type for detecting bit errors.  Default value is 0(NO_PARITY).
    BIT_SAMPLE_POINT = 0x17  # 0-100	# CAN specific, the desired bit sample point as a percentage of bit time. Default value is 80%.
    SYNCH_JUMP_WIDTH = 0x18  # 0-100	# CAN specific, the desired synchronization jump width as a percentage of the bit time. Default value is 15%.
    W0 = 0x19
    T1_MAX = 0x1A  # 0x0-0xFFFF	# SCI_X_XXXX specific, the max. interframe response delay. Default value is 20 ms.
    T2_MAX = 0x1B  # 0x0-0xFFFF	# SCI_X_XXXX specific, the max. interframe request delay.Default value is 100 ms.
    T4_MAX = 0x1C  # 0x0-0xFFFF	# SCI_X_XXXX specific, the max. intermessage response delay. Default value is 20 ms.
    T5_MAX = 0x1D  # 0x0-0xFFFF	# SCI_X_XXXX specific, the max. intermessage request delay. Default value is 100 ms.
    ISO15765_BS = 0x1E  # 0x0-0xFF	# ISO15765 specific, the block size for segmented transfers.
    ISO15765_STMIN = 0x1F  # 0x0-0xFF	# ISO15765 specific, the separation time for segmented transfers.
    DATA_BITS = 0x20  # 04.04-API only
    FIVE_BAUD_MOD = 0x21
    BS_TX = 0x22
    STMIN_TX = 0x23
    T3_MAX = 0x24
    ISO15765_WFT_MAX = 0x25


class Ioctl_Flags(Enum):
    TX_IOCTL_BASE = 0x70000
    TX_IOCTL_SET_DLL_DEBUG_FLAGS = 0x70001
    TX_IOCTL_DLL_DEBUG_FLAG_J2534_CALLS = 0x00000001



import struct
from udsoncan import DataFormatIdentifier, MemoryLocation
from udsoncan.Request import Request
from udsoncan.Response import Response
from udsoncan.exceptions import *
from udsoncan.BaseService import BaseService, BaseResponseData
from udsoncan.ResponseCode import ResponseCode

from typing import Optional, cast


class RequestUpload(BaseService):
    _sid = 0x35
    _use_subfunction = False

    supported_negative_response = [ResponseCode.IncorrectMessageLengthOrInvalidFormat,
                                   ResponseCode.ConditionsNotCorrect,
                                   ResponseCode.RequestOutOfRange,
                                   ResponseCode.SecurityAccessDenied,
                                   ResponseCode.UploadDownloadNotAccepted
                                   ]

    class ResponseData(BaseResponseData):
        """
        .. data:: max_length

                (int) Maximum number of data blocks to read
        """

        max_length: int

        def __init__(self, max_length):
            super().__init__(RequestUpload)
            self.max_length = max_length

    class InterpretedResponse(Response):
        service_data: "RequestUpload.ResponseData"

    @classmethod
    def normalize_data_format_identifier(cls, dfi: Optional[DataFormatIdentifier]) -> DataFormatIdentifier:
        if dfi is None:
            dfi = DataFormatIdentifier()

        if not isinstance(dfi, DataFormatIdentifier):
            raise ValueError('dfi must be an instance of DataFormatIdentifier')

        return dfi

    @classmethod
    def make_request(cls, memory_location: MemoryLocation, dfi: Optional[DataFormatIdentifier] = None) -> Request:
        """
        Generates a request for RequestUpload

        :param memory_location: The address and the size of the memory block to be read.
        :type memory_location: :ref:`MemoryLocation <MemoryLocation>`

        :param dfi: Optional dataFormatIdentifier defining the compression and encryption scheme of the data. 
                If not specified, the default value of 00 will be used, specifying no encryption and no compression
        :type dfi: :ref:`DataFormatIdentifier <DataFormatIdentifier>`		

        :raises ValueError: If parameters are out of range, missing or wrong type
        """

        dfi = cls.normalize_data_format_identifier(dfi)

        if not isinstance(memory_location, MemoryLocation):
            raise ValueError('memory_location must be an instance of MemoryLocation')

        request = Request(service=cls)
        request.data = bytes()
        request.data += dfi.get_byte()  # Data Format Identifier
        request.data += memory_location.alfid.get_byte()  # AddressAndLengthFormatIdentifier
        request.data += memory_location.get_address_bytes()
        request.data += memory_location.get_memorysize_bytes()

        return request

    @classmethod
    def interpret_response(cls, response: Response) -> InterpretedResponse:
        """
        Populates the response ``service_data`` property with an instance of :class:`RequestUpload.ResponseData<udsoncan.services.RequestUpload.ResponseData>`

        :param response: The received response to interpret
        :type response: :ref:`Response<Response>`

        :raises InvalidResponseException: If length of ``response.data`` is too short
        :raises NotImplementedError: If the maxNumberOfBlockLength value is encoded over more than 8 bytes.
        """
        if response.data is None:
            raise InvalidResponseException(response, "No data in response")

        if len(response.data) < 1:
            raise InvalidResponseException(response, "Response data must be at least 1 bytes")

        lfid = int(response.data[0]) >> 4

        if lfid > 8:
            raise NotImplementedError('This client does not support number bigger than %d bits' % (8 * 8))

        if len(response.data) < lfid + 1:
            raise InvalidResponseException(
                response, "Length of data (%d) is too short to contains the number of block of given length (%d)" % (len(response.data), lfid))

        todecode = bytearray(b'\x00\x00\x00\x00\x00\x00\x00\x00')
        for i in range(1, lfid + 1):
            todecode[-i] = response.data[lfid + 1 - i]

        response.service_data = cls.ResponseData(
            max_length=struct.unpack('>q', todecode)[0]
        )

        return cast(RequestUpload.InterpretedResponse, response)



import struct
from typing import Optional, cast

from udsoncan.Request import Request
from udsoncan.Response import Response
from udsoncan.exceptions import *
from udsoncan.BaseService import BaseService, BaseResponseData, BaseSubfunction
from udsoncan.ResponseCode import ResponseCode
import udsoncan.tools as tools


class AuthenticationReturnParameter:
    RequestAccepted = 0x00
    GeneralReject = 0x01
    AuthenticationConfiguration_APCE = 0x02
    AuthenticationConfiguration_ACR_with_asymmetric_cryptography = 0x03
    AuthenticationConfiguration_ACR_with_symmetric_cryptography = 0x04

    DeAuthentication_successful = 0x10
    CertificateVerified_OwnershipVerificationNecessary = 0x11
    OwnershipVerified_AuthenticationComplete = 0x12
    CertificateVerified = 0x13


class Authentication(BaseService):
    _sid = 0x29

    supported_negative_response = [ResponseCode.SubFunctionNotSupported,
                                   ResponseCode.IncorrectMessageLengthOrInvalidFormat,
                                   ResponseCode.ConditionsNotCorrect,
                                   ResponseCode.RequestSequenceError,
                                   ResponseCode.CertificateVerificationFailed_InvalidTimePeriod,
                                   ResponseCode.CertificateVerificationFailed_InvalidSignature,
                                   ResponseCode.CertificateVerificationFailed_InvalidChainOfTrust,
                                   ResponseCode.CertificateVerificationFailed_InvalidType,
                                   ResponseCode.CertificateVerificationFailed_InvalidFormat,
                                   ResponseCode.CertificateVerificationFailed_InvalidContent,
                                   ResponseCode.CertificateVerificationFailed_InvalidScope,
                                   ResponseCode.CertificateVerificationFailed_InvalidCertificate,
                                   ResponseCode.OwnershipVerificationFailed,
                                   ResponseCode.ChallengeCalculationFailed,
                                   ResponseCode.SettingAccessRightsFailed,
                                   ResponseCode.SessionKeyCreationDerivationFailed,
                                   ResponseCode.ConfigurationDataUsageFailed,
                                   ResponseCode.DeAuthenticationFailed]

    class ResponseData(BaseResponseData):
        """
        .. data:: authentication_task_echo

            Authentication Task echoed back by the server

        .. data:: return_value

            :class:`AuthenticationReturnParameter<AuthenticationReturnParameter>` as defined in
             B.5. (ISO 14229-1:2020)

        .. data:: challenge_server

            The challenge contains vehicle manufacturer specific formatted server data (eventually containing randomized
            information) or is a random number.

        .. data:: ephemeral_public_key_server

            Ephemeral public key generated by the server for Diffie-Hellman key agreement.

        .. data:: certificate_server

            The Certificate to verify.

        .. data:: proof_of_ownership_server

            Proof of Ownership to be verified by the client.

        .. data:: session_key_info

            If present, this value shall contain session key information, e.g. the encrypted session key(s) for securing
            further communication in the actual session and/or proof value(s) (e.g. a hash value) for the validation of the
            session key(s) on the client side.

        .. data:: algorithm_indicator

            Indicates the algorithm used in the generating and verifying Proof of Ownership (POWN), which further determines
            the parameters used in the algorithm and possibly the session key creation mode. This field is a 16 byte value
            containing the BER encoded OID value of the algorithm used. The value is left aligned and right padded with
            zero up to 16 bytes.

        .. data:: needed_additional_parameter

            Indicate what additional parameters, if needed, are expected by the server.
        """
        authentication_task_echo: int
        return_value: int
        challenge_server: Optional[bytes]
        ephemeral_public_key_server: Optional[bytes]
        certificate_server: Optional[bytes]
        proof_of_ownership_server: Optional[bytes]
        session_key_info: Optional[bytes]
        algorithm_indicator: Optional[bytes]
        needed_additional_parameter: Optional[bytes]

        def __init__(self, authentication_task_echo: int,
                     return_value: int,
                     challenge_server: Optional[bytes] = None,
                     ephemeral_public_key_server: Optional[bytes] = None,
                     certificate_server: Optional[bytes] = None,
                     proof_of_ownership_server: Optional[bytes] = None,
                     session_key_info: Optional[bytes] = None,
                     algorithm_indicator: Optional[bytes] = None,
                     needed_additional_parameter: Optional[bytes] = None
                     ):
            super().__init__(Authentication)
            self.authentication_task_echo = authentication_task_echo
            self.return_value = return_value
            self.challenge_server = challenge_server
            self.ephemeral_public_key_server = ephemeral_public_key_server
            self.certificate_server = certificate_server
            self.session_key_info = session_key_info
            self.proof_of_ownership_server = proof_of_ownership_server
            self.algorithm_indicator = algorithm_indicator
            self.needed_additional_parameter = needed_additional_parameter

    class InterpretedResponse(Response):
        service_data: "Authentication.ResponseData"

    class AuthenticationTask(BaseSubfunction):
        __pretty_name__ = 'authentication task'

        deAuthenticate = 0
        verifyCertificateUnidirectional = 1
        verifyCertificateBidirectional = 2
        proofOfOwnership = 3
        transmitCertificate = 4
        requestChallengeForAuthentication = 5
        verifyProofOfOwnershipUnidirectional = 6
        verifyProofOfOwnershipBidirectional = 7
        authenticationConfiguration = 8

    @staticmethod
    def _append_byes_parameter(data: bytes, parameter: Optional[bytes], parameter_name: str) -> bytes:
        if parameter is not None:

            if not isinstance(parameter, bytes):
                raise ValueError(f'{parameter_name} must be a bytes object')

            tools.validate_int(len(parameter), min=0, max=0xFFFF, name=f'{parameter_name} length')
            data += struct.pack('>H', len(parameter))
            data += parameter
        else:
            data += struct.pack('>H', 0)

        return data

    @classmethod
    def make_request(cls, authentication_task: int,
                     communication_configuration: Optional[int] = None,
                     certificate_client: Optional[bytes] = None,
                     challenge_client: Optional[bytes] = None,
                     algorithm_indicator: Optional[bytes] = None,
                     certificate_evaluation_id: Optional[int] = None,
                     certificate_data: Optional[bytes] = None,
                     proof_of_ownership_client: Optional[bytes] = None,
                     ephemeral_public_key_client: Optional[bytes] = None,
                     additional_parameter: Optional[bytes] = None) -> Request:
        """
        Generates a request for Authentication

        :param authentication_task: Service subfunction. Allowed values are from 0 to 8.
        :type authentication_task: int

        :param communication_configuration: Optional Configuration information about how to proceed with security
          in further diagnostic communication after the Authentication (vehicle manufacturer specific).
          Allowed values are from 0 to 255.
        :type communication_configuration: int

        :param certificate_client: Optional The Certificate to verify.
        :type certificate_client: bytes or None

        :param challenge_client: Optional The challenge contains vehicle manufacturer specific formatted client data
          (likely containing randomized information) or is a random number.
        :type challenge_client: bytes or None

        :param algorithm_indicator: Optional Indicates the algorithm used in the generating and verifying Proof of
          Ownership (POWN), which further determines the parameters used in the algorithm and possibly the session key
          creation mode. This field is a 16 byte value containing the BER encoded OID value of the algorithm used.
          The value is left aligned and right padded with zero up to 16 bytes.
        :type algorithm_indicator: bytes or None

        :param certificate_evaluation_id: Optional unique ID to identify the evaluation type of the transmitted
          certificate. The value of this parameter is vehicle manufacturer specific. Subsequent diagnostic requests with
          the same evaluationTypeId will overwrite the certificate data of the previous requests.
          Allowed values are from 0 to 0xFFFF.
        :type certificate_evaluation_id: int or None

        :param certificate_data: Optional The Certificate to verify.
        :type certificate_data: bytes or None

        :param proof_of_ownership_client: Optional Proof of Ownership of the previous given challenge to be verified by
          the server.
        :type proof_of_ownership_client: bytes or None

        :param ephemeral_public_key_client: Optional Ephemeral public key generated by the client for Diffie-Hellman key
          agreement.
        :type ephemeral_public_key_client: bytes or None

        :param additional_parameter: Optional additional parameter is provided to the server if the server indicates
          as neededAdditionalParameter.
        :type additional_parameter: bytes or None

        :raises ValueError: If parameters are out of range, missing or wrong type
        """

        tools.validate_int(authentication_task, min=0, max=8, name='Authentication Task')
        data = None
        if authentication_task in (Authentication.AuthenticationTask.deAuthenticate,
                                   Authentication.AuthenticationTask.authenticationConfiguration):
            # Service ID + Subfunction, no param
            pass
        elif authentication_task in (Authentication.AuthenticationTask.verifyCertificateUnidirectional,
                                     Authentication.AuthenticationTask.verifyCertificateBidirectional,
                                     Authentication.AuthenticationTask.requestChallengeForAuthentication):
            tools.validate_int(communication_configuration, min=0, max=0xFF, name='Communication Configuration')
            # communicationConfiguration
            data = struct.pack('B', communication_configuration)
            if authentication_task in (Authentication.AuthenticationTask.verifyCertificateUnidirectional,
                                       Authentication.AuthenticationTask.verifyCertificateBidirectional):
                # certificateClient
                data = Authentication._append_byes_parameter(data, certificate_client, 'Certificate Client')

                # challengeClient
                data = Authentication._append_byes_parameter(data, challenge_client, 'Challenge Client')
            else:
                if not isinstance(algorithm_indicator, bytes) or len(algorithm_indicator) != 16:
                    raise ValueError(f'{algorithm_indicator!r} must be a bytes object of length 16')
                # algorithmIndicator
                data += algorithm_indicator
        elif authentication_task == Authentication.AuthenticationTask.proofOfOwnership:
            data = b''
            # proofOfOwnershipClient
            data = Authentication._append_byes_parameter(data, proof_of_ownership_client, 'Proof Of Ownership Client')
            # ephemeralPublicKeyClient
            data = Authentication._append_byes_parameter(data, ephemeral_public_key_client, 'Ephemeral Public Key'
                                                                                            ' Client')

        elif authentication_task == Authentication.AuthenticationTask.transmitCertificate:
            tools.validate_int(certificate_evaluation_id, min=0, max=0xFFFF, name='Certificate Evaluation ID')
            # certificateEvaluationId
            data = struct.pack('>H', certificate_evaluation_id)
            # certificateData
            data = Authentication._append_byes_parameter(data, certificate_data, 'Certificate Data')

        elif authentication_task in (Authentication.AuthenticationTask.verifyProofOfOwnershipUnidirectional,
                                     Authentication.AuthenticationTask.verifyProofOfOwnershipBidirectional):
            if not isinstance(algorithm_indicator, bytes) or len(algorithm_indicator) != 16:
                raise ValueError(f'{algorithm_indicator!r} must be a bytes object of length 16')
            # algorithmIndicator
            data = algorithm_indicator
            # proofOfOwnershipClient
            data = Authentication._append_byes_parameter(data, proof_of_ownership_client, 'Proof Of Ownership Client')
            # challengeClient
            data = Authentication._append_byes_parameter(data, challenge_client, 'Challenge Client')
            # additionalParameter
            data = Authentication._append_byes_parameter(data, additional_parameter, 'Additional Parameter')

        return Request(service=cls, subfunction=authentication_task, data=data)

    @staticmethod
    def _extract_byes_parameter(response: Response, data: bytes, parameter_name: str) -> bytes:
        if data is not None and len(data) >= 2:
            parameter_length, = struct.unpack('>H', data[:2])
        else:
            raise InvalidResponseException(response, f'Not enough data for length of {parameter_name}')

        if len(data) >= (2 + parameter_length):
            return data[2:(2 + parameter_length)]
        else:
            raise InvalidResponseException(response, f'Not enough data for {parameter_name}.'
                                                     f' Parsed length is: {parameter_length}, reset of payload length'
                                                     f' is {len(data) - 2}')

    @classmethod
    def interpret_response(cls, response: Response) -> InterpretedResponse:
        """
        Populates the response ``service_data`` property with an instance of
         :class:`Authentication.ResponseData<udsoncan.services.Authentication.ResponseData>`

        :param response: The received response to interpret
        :type response: :ref:`Response<Response>`

        :raises InvalidResponseException: If response data length is not as expected
        """

        if response.data is None:
            raise InvalidResponseException(response, "No data in response")

        if len(response.data) < 2:
            raise InvalidResponseException(response, "Response data must be at least 2 bytes")

        subfunction = response.data[0]
        return_value = response.data[1]
        offset = 2
        challenge_server: Optional[bytes] = None
        ephemeral_public_key_server: Optional[bytes] = None
        certificate_server: Optional[bytes] = None
        proof_of_ownership_server: Optional[bytes] = None
        session_key_info: Optional[bytes] = None
        algorithm_indicator: Optional[bytes] = None
        needed_additional_parameter: Optional[bytes] = None

        if subfunction in (Authentication.AuthenticationTask.deAuthenticate,
                           Authentication.AuthenticationTask.transmitCertificate,
                           Authentication.AuthenticationTask.authenticationConfiguration):
            pass  # no more parameters

        elif subfunction == Authentication.AuthenticationTask.verifyCertificateUnidirectional:
            challenge_server = Authentication._extract_byes_parameter(response, response.data[offset:],
                                                                      'Challenge Server')
            offset += 2 + len(challenge_server)

            ephemeral_public_key_server = Authentication._extract_byes_parameter(response, response.data[offset:],
                                                                                 'Ephemeral Public Key Server')
            offset += 2 + len(ephemeral_public_key_server)

        elif subfunction == Authentication.AuthenticationTask.verifyCertificateBidirectional:
            challenge_server = Authentication._extract_byes_parameter(response, response.data[offset:],
                                                                      'Challenge Server')
            offset += 2 + len(challenge_server)

            certificate_server = Authentication._extract_byes_parameter(response, response.data[offset:],
                                                                        'Certificate Server')
            offset += 2 + len(certificate_server)

            proof_of_ownership_server = Authentication._extract_byes_parameter(response, response.data[offset:],
                                                                               'Proof Of Ownership Server')
            offset += 2 + len(proof_of_ownership_server)

            ephemeral_public_key_server = Authentication._extract_byes_parameter(response, response.data[offset:],
                                                                                 'Ephemeral Public Key Server')
            offset += 2 + len(ephemeral_public_key_server)

        elif subfunction == Authentication.AuthenticationTask.proofOfOwnership:
            session_key_info = Authentication._extract_byes_parameter(response, response.data[offset:],
                                                                      'Session Key Info')
            offset += 2 + len(session_key_info)

        elif subfunction in (Authentication.AuthenticationTask.requestChallengeForAuthentication,
                             Authentication.AuthenticationTask.verifyProofOfOwnershipUnidirectional,
                             Authentication.AuthenticationTask.verifyProofOfOwnershipBidirectional
                             ):
            if len(response.data[offset:]) < 16:  # algorithmIndicator is exactly 16 bytes
                raise InvalidResponseException(response, f'Not enough data for Algorithm Indicator (16 bytes), reset of'
                                                         f' payload length is {len(response.data) - offset}')
            algorithm_indicator = response.data[offset:(offset + 16)]
            offset += len(algorithm_indicator)

            if subfunction == Authentication.AuthenticationTask.requestChallengeForAuthentication:
                challenge_server = Authentication._extract_byes_parameter(response, response.data[offset:],
                                                                          'Challenge Server')
                offset += 2 + len(challenge_server)

                needed_additional_parameter = Authentication._extract_byes_parameter(response, response.data[offset:],
                                                                                     'Needed Additional Parameter')
                offset += 2 + len(needed_additional_parameter)
            else:
                if subfunction == Authentication.AuthenticationTask.verifyProofOfOwnershipBidirectional:
                    proof_of_ownership_server = Authentication._extract_byes_parameter(response, response.data[offset:],
                                                                                       'Proof Of Ownership Server')
                    offset += 2 + len(proof_of_ownership_server)

                # verifyProofOfOwnershipBidirectional or verifyProofOfOwnershipUnidirectional
                session_key_info = Authentication._extract_byes_parameter(response, response.data[offset:],
                                                                          'Session Key Info')
                offset += 2 + len(session_key_info)
        else:
            raise InvalidResponseException(response, "Unknown Subfunction")

        if len(response.data) > offset:
            raise InvalidResponseException(response, "Response data had more bytes than expected")

        response.service_data = cls.ResponseData(
            authentication_task_echo=subfunction,
            return_value=return_value,
            challenge_server=challenge_server,
            ephemeral_public_key_server=ephemeral_public_key_server,
            certificate_server=certificate_server,
            proof_of_ownership_server=proof_of_ownership_server,
            session_key_info=session_key_info,
            algorithm_indicator=algorithm_indicator,
            needed_additional_parameter=needed_additional_parameter
        )

        return cast(Authentication.InterpretedResponse, response)



import logging
import logging.config
from os import path

from udsoncan.exceptions import *
from udsoncan.Response import Response
from udsoncan.Request import Request

from udsoncan.common.AddressAndLengthFormatIdentifier import *
from udsoncan.common.Baudrate import *
from udsoncan.common.CommunicationType import *
from udsoncan.common.DataFormatIdentifier import *
from udsoncan.common.dids import *
from udsoncan.common.DidCodec import *
from udsoncan.common.dtc import *
from udsoncan.common.DynamicDidDefinition import *
from udsoncan.common.Filesize import *
from udsoncan.common.IOControls import *
from udsoncan.common.MemoryLocation import *
from udsoncan.common.Routine import *
from udsoncan.common.Units import *
from udsoncan.typing import *

__version__ = '1.25.0'
__license__ = 'MIT'
__author__ = 'Pier-Yves Lessard'

latest_standard = 2020
__default_log_config_file = path.join(path.dirname(path.abspath(__file__)), 'logging.conf')


def setup_logging(config_file=__default_log_config_file):
    """
    This function setup the logger accordingly to the module provided cfg file
    """
    try:
        logging.config.fileConfig(config_file)
    except Exception as e:
        logging.warning('Cannot load logging configuration from %s. %s:%s' % (config_file, e.__class__.__name__, str(e)))



__all__ = ['MemoryLocation']

import struct
import math

from udsoncan.common.AddressAndLengthFormatIdentifier import AddressAndLengthFormatIdentifier

from typing import Optional


class MemoryLocation:
    """
    This class defines a memory block location including : address, size, AddressAndLengthFormatIdentifier (address format and memory size format)

    :param address: A memory address pointing to the beginning of the memory block
    :type address: int

    :param memorysize: The size of the memory block
    :type memorysize: int

    :param address_format: The number of bits on which an address should be encoded. Possible values are 8, 16, 24, 32, 40.
            If ``None`` is specified, the smallest size required to store the given address will be used
    :type address_format: int or None

    :param memorysize_format: The number of bits on which a memory size should be encoded. Possible values are 8, 16, 24, 32
            If ``None`` is specified, the smallest size required to store the given memorysize will be used
    :type memorysize_format: int or None	

    """

    address: int
    memorysize: int
    address_format: Optional[int]
    memorysize_format: Optional[int]

    def __init__(self, address: int, memorysize: int, address_format: Optional[int] = None, memorysize_format: Optional[int] = None):
        self.address = address
        self.memorysize = memorysize
        self.address_format = address_format
        self.memorysize_format = memorysize_format

        if address_format is None:
            address_format = self.autosize_address(address)

        if memorysize_format is None:
            memorysize_format = self.autosize_memorysize(memorysize)

        self.alfid = AddressAndLengthFormatIdentifier(memorysize_format=memorysize_format, address_format=address_format)

    # This is used by the client/server to set a format from a config object while letting the user override it
    def set_format_if_none(self, address_format: Optional[int] = None, memorysize_format: Optional[int] = None) -> None:
        previous_address_format = self.address_format
        previous_memorysize_format = self.memorysize_format
        try:
            if address_format is not None:
                if self.address_format is None:
                    self.address_format = address_format

            if memorysize_format is not None:
                if self.memorysize_format is None:
                    self.memorysize_format = memorysize_format

            address_format = self.address_format if self.address_format is not None else self.autosize_address(self.address)
            memorysize_format = self.memorysize_format if self.memorysize_format is not None else self.autosize_memorysize(self.memorysize)

            self.alfid = AddressAndLengthFormatIdentifier(memorysize_format=memorysize_format, address_format=address_format)
        except:
            self.address_format = previous_address_format
            self.memorysize_format = previous_memorysize_format
            raise

    # Finds the smallest size that fits the address
    def autosize_address(self, val: int) -> int:
        fmt = math.ceil(val.bit_length() / 8) * 8
        if fmt > 64:
            raise ValueError("address size must be smaller or equal than 64 bits")
        return fmt

    # Finds the smallest size that fits the memory size
    def autosize_memorysize(self, val: int) -> int:
        fmt = math.ceil(val.bit_length() / 8) * 8
        if fmt > 64:
            raise ValueError("memory size must be smaller or equal than 64 bits")
        return fmt

    # Gets the address byte in the requested format
    def get_address_bytes(self) -> bytes:
        n = AddressAndLengthFormatIdentifier.address_map[self.alfid.address_format]

        data = struct.pack('>q', self.address)
        return data[-n:]

    # Gets the memory size byte in the requested format

    def get_memorysize_bytes(self) -> bytes:
        n = AddressAndLengthFormatIdentifier.memsize_map[self.alfid.memorysize_format]

        data = struct.pack('>q', self.memorysize)
        return data[-n:]

    # Generates an instance from the byte stream
    @classmethod
    def from_bytes(cls, address_bytes: bytes, memorysize_bytes: bytes) -> "MemoryLocation":
        if not isinstance(address_bytes, bytes):
            raise ValueError('address_bytes must be a valid bytes object')

        if not isinstance(memorysize_bytes, bytes):
            raise ValueError('memorysize_bytes must be a valid bytes object')

        if len(address_bytes) > 5:
            raise ValueError('Address must be at most 40 bits long')

        if len(memorysize_bytes) > 4:
            raise ValueError('Memory size must be at most 32 bits long')

        address_bytes_padded = b'\x00' * (8 - len(address_bytes)) + address_bytes
        memorysize_bytes_padded = b'\x00' * (8 - len(memorysize_bytes)) + memorysize_bytes

        address = struct.unpack('>q', address_bytes_padded)[0]
        memorysize = struct.unpack('>q', memorysize_bytes_padded)[0]
        address_format = len(address_bytes) * 8
        memorysize_format = len(memorysize_bytes) * 8

        return cls(address=address, memorysize=memorysize, address_format=address_format, memorysize_format=memorysize_format)

    def __str__(self) -> str:
        return 'Address=0x%x (%d bits), Size=0x%x (%d bits)' % (self.address, self.alfid.address_format, self.memorysize, self.alfid.memorysize_format)

    def __repr__(self) -> str:
        return '<%s: %s at 0x%08x>' % (self.__class__.__name__, str(self), id(self))



import struct
from udsoncan import Dtc, check_did_config, make_did_codec_from_definition, fetch_codec_definition_from_config, latest_standard, DIDConfig
from udsoncan.Request import Request
from udsoncan.Response import Response
from udsoncan.exceptions import *
from udsoncan.BaseService import BaseService, BaseSubfunction, BaseResponseData
from udsoncan.ResponseCode import ResponseCode
import udsoncan.tools as tools

from typing import Dict, Union, Optional, cast, List


class ReadDTCInformation(BaseService):
    _sid = 0x19

    supported_negative_response = [ResponseCode.SubFunctionNotSupported,
                                   ResponseCode.IncorrectMessageLengthOrInvalidFormat,
                                   ResponseCode.RequestOutOfRange
                                   ]

    class ResponseData(BaseResponseData):
        """
        .. data:: subfunction_echo

                Subfunction echoed back by the server

        .. data:: dtcs

                :ref:`DTC<DTC>` instances and their status read from the server.

        .. data:: dtc_count

                Number of DTC read or available

        .. data:: dtc_format

                Integer indicating the format of the DTC. See :ref:`DTC.Format<DTC_Format>`

        .. data:: status_availability

                :ref:`Dtc.Status<DTC_Status>` indicating which status the server supports

        .. data:: extended_data

                List of bytes containing the DTC extended data

        .. data:: memory_selection_echo

                Echo of the memory selection byte                               

        """

        subfunction_echo: int
        dtcs: List[Dtc]
        dtc_count: Optional[int]
        dtc_format: Optional[int]
        status_availability: Optional[Dtc.Status]
        extended_data: Optional[List[bytes]]
        memory_selection_echo: Optional[int]
        functional_group_id:Optional[int]
        severity_availability: Optional[Dtc.Severity]

        def __init__(self, subfunction_echo: int,
                     dtcs: Optional[List[Dtc]] = None,
                     dtc_count: Optional[int] = None,
                     dtc_format: Optional[int] = None,
                     status_availability: Optional[Dtc.Status] = None,
                     extended_data: Optional[List[bytes]] = [],
                     memory_selection_echo: Optional[int] = None,
                     functional_group_id:Optional[int] = None,
                     severity_availability: Optional[Dtc.Severity] = None
                     ):
            super().__init__(ReadDTCInformation)
            self.subfunction_echo = subfunction_echo
            self.dtcs = dtcs if dtcs is not None else []
            self.dtc_count = dtc_count
            self.dtc_format = dtc_format
            self.status_availability = status_availability
            self.extended_data = extended_data if extended_data is not None else []
            self.memory_selection_echo = memory_selection_echo
            self.functional_group_id = functional_group_id
            self.severity_availability=severity_availability

    class InterpretedResponse(Response):
        service_data: "ReadDTCInformation.ResponseData"

    class Subfunction(BaseSubfunction):
        __pretty_name__ = 'subfunction'

        reportNumberOfDTCByStatusMask = 1
        reportDTCByStatusMask = 2
        reportDTCSnapshotIdentification = 3
        reportDTCSnapshotRecordByDTCNumber = 4
        reportDTCSnapshotRecordByRecordNumber = 5
        reportDTCExtendedDataRecordByDTCNumber = 6
        reportNumberOfDTCBySeverityMaskRecord = 7
        reportDTCBySeverityMaskRecord = 8
        reportSeverityInformationOfDTC = 9
        reportSupportedDTCs = 0xA
        reportFirstTestFailedDTC = 0xB
        reportFirstConfirmedDTC = 0xC
        reportMostRecentTestFailedDTC = 0xD
        reportMostRecentConfirmedDTC = 0xE
        reportMirrorMemoryDTCByStatusMask = 0xF
        reportMirrorMemoryDTCExtendedDataRecordByDTCNumber = 0x10
        reportNumberOfMirrorMemoryDTCByStatusMask = 0x11
        reportNumberOfEmissionsRelatedOBDDTCByStatusMask = 0x12
        reportEmissionsRelatedOBDDTCByStatusMask = 0x13
        reportDTCFaultDetectionCounter = 0x14
        reportDTCWithPermanentStatus = 0x15
        reportDTCExtDataRecordByRecordNumber = 0x16
        reportUserDefMemoryDTCByStatusMask = 0x17
        reportUserDefMemoryDTCSnapshotRecordByDTCNumber = 0x18
        reportUserDefMemoryDTCExtDataRecordByDTCNumber = 0x19
        reportWWHOBDDTCByMaskRecord = 0x42
        reportWWHOBDDTCWithPermanentStatus = 0x55

        # todo
        reportSupportedDTCExtDataRecord = 0x1A
        reportDTCInformationByDTCReadinessGroupIdentifier = 0x56

    @classmethod
    def assert_severity_mask(cls, severity_mask: Optional[int], subfunction: int) -> None:
        if severity_mask is None:
            raise ValueError('severity_mask must be provided for subfunction 0x%02x' % subfunction)
        tools.validate_int(severity_mask, min=0, max=0xFF, name='Severity mask')

    @classmethod
    def assert_status_mask(cls, status_mask: Optional[int], subfunction: int) -> None:
        if status_mask is None:
            raise ValueError('status_mask must be provided for subfunction 0x%02x' % subfunction)
        tools.validate_int(status_mask, min=0, max=0xFF, name='Status mask')

    @classmethod
    def assert_memory_selection(cls, memory_selection: Optional[int], subfunction: int) -> None:
        if memory_selection is None:
            raise ValueError('memory_selection must be provided for subfunction 0x%02x' % subfunction)
        tools.validate_int(memory_selection, min=0, max=0xFF, name='Memory Selection')

    @classmethod
    def assert_dtc(cls, dtc: Optional[int], subfunction: int) -> None:
        if dtc is None:
            raise ValueError('A dtc value must be provided for subfunction 0x%02x' % subfunction)
        tools.validate_int(dtc, min=0, max=0xFFFFFF, name='DTC')

    @classmethod
    def assert_snapshot_record_number(cls, snapshot_record_number: Optional[int], subfunction: int) -> None:
        if snapshot_record_number is None:
            raise ValueError('snapshot_record_number must be provided for subfunction 0x%02x' % subfunction)
        tools.validate_int(snapshot_record_number, min=0, max=0xFF, name='Snapshot record number')

    @classmethod
    def assert_extended_data_record_number(cls, extended_data_record_number: Optional[int], subfunction: int, maxval: int = 0xFF) -> None:
        if extended_data_record_number is None:
            raise ValueError('extended_data_record_number must be provided for subfunction 0x%02x' % subfunction)
        tools.validate_int(extended_data_record_number, min=0, max=maxval, name='Extended data record number')

    @classmethod
    def assert_extended_data_size_int_or_dict(cls,
                                              extended_data_size: Optional[Union[int, Dict[int, int]]],
                                              subfunction: int) -> None:
        if extended_data_size is None:
            raise ValueError('extended_data_size must be provided as length of data is not given by the server for subfunction %d.' % subfunction)
        if isinstance(extended_data_size, int):
            tools.validate_int(extended_data_size, min=0, max=0xFFF, name='Extended data size')
        if isinstance(extended_data_size, dict):
            for dtcid in extended_data_size:
                tools.validate_int(extended_data_size[dtcid], min=0, max=0xFFF, name='Extended data size for DTC=0x%06x' % dtcid)

    @classmethod
    def assert_functional_group_id(cls, functional_group_id: Optional[int], subfunction) -> None:
        if functional_group_id is None:
            raise ValueError('functional_group_id must be provided for subfunction 0x%02x' % subfunction)
        tools.validate_int(functional_group_id, min=0, max=0xFE, name='Functional Group ID')

    @classmethod
    def pack_dtc(cls, dtcid: int) -> bytes:
        return struct.pack('BBB', (dtcid >> 16) & 0xFF, (dtcid >> 8) & 0xFF, (dtcid >> 0) & 0xFF)

    @classmethod
    def check_subfunction_valid(cls, subfunction: int, standard_version: int = latest_standard) -> None:
        tools.validate_int(subfunction, min=1, max=0xFF, name='Subfunction')
        vlist = vars(cls)
        ok = True
        for v in vlist:
            if isinstance(v, int) and v == subfunction:
                ok = False
                break
        if not ok:
            raise ValueError('Unknown subfunction : 0x%02x', subfunction)

        # These subfunction have been added in the 2020 version of the standard
        subfunction2020 = [
            cls.Subfunction.reportUserDefMemoryDTCByStatusMask,
            cls.Subfunction.reportDTCExtDataRecordByRecordNumber,
            cls.Subfunction.reportUserDefMemoryDTCSnapshotRecordByDTCNumber,
            cls.Subfunction.reportUserDefMemoryDTCExtDataRecordByDTCNumber,
            cls.Subfunction.reportSupportedDTCExtDataRecord,
            cls.Subfunction.reportWWHOBDDTCByMaskRecord,
            cls.Subfunction.reportWWHOBDDTCWithPermanentStatus,
            cls.Subfunction.reportDTCInformationByDTCReadinessGroupIdentifier,
        ]

        if subfunction in subfunction2020 and standard_version < 2020:
            raise NotImplementedError('The subfunction 0x%02x has been introduced in standard version 2020 but decoding is requested to be done as per %d version' % (
                subfunction, standard_version))

    @classmethod
    def make_request(cls,
                     subfunction: int,
                     status_mask: Optional[Union[int, Dtc.Status]] = None,
                     severity_mask: Optional[Union[int, Dtc.Severity]] = None,
                     dtc_class: Optional[Union[int, Dtc.DtcClass]] = None,
                     dtc: Optional[Union[int, Dtc]] = None,
                     snapshot_record_number: Optional[int] = None,
                     extended_data_record_number: Optional[int] = None,
                     memory_selection: Optional[int] = None,
                     standard_version: int = latest_standard,
                     functional_group_id:Optional[int] = None,
                     ) -> Request:
        """
        Generates a request for ReadDTCInformation. 
        Each subfunction uses a subset of parameters. 

        :param subfunction: The service subfunction. Values are defined in :class:`ReadDTCInformation.Subfunction<ReadDTCInformation.Subfunction>`
        :type subfunction: int

        :param status_mask: A DTC status mask used to filter DTC
        :type status_mask: int or :ref:`Dtc.Status <DTC_Status>`

        :param severity_mask: A severity mask used to filter DTC 
        :type severity_mask: int or :ref:`Dtc.Severity <DTC_Severity>`

        :param dtc_class: A DTC class to be packed in the same byte as the severity mask
        :type dtc_class: int or :ref:`Dtc.DtcClass <DTC_DtcClass>`

        :param dtc: A DTC mask used to filter DTC
        :type dtc: int or :ref:`Dtc <DTC>`

        :param snapshot_record_number: Snapshot record number
        :type snapshot_record_number: int

        :param extended_data_record_number: Extended data record number
        :type extended_data_record_number: int

        :param memory_selection: Memory selection for user defined memory DTC
        :type memory_selection: int


        :raises ValueError: If parameters are out of range, missing or wrong type
        """

        # Request grouping for subfunctions that have the same request format
        request_subfn_no_param = [
            ReadDTCInformation.Subfunction.reportSupportedDTCs,
            ReadDTCInformation.Subfunction.reportFirstTestFailedDTC,
            ReadDTCInformation.Subfunction.reportFirstConfirmedDTC,
            ReadDTCInformation.Subfunction.reportMostRecentTestFailedDTC,
            ReadDTCInformation.Subfunction.reportMostRecentConfirmedDTC,
            ReadDTCInformation.Subfunction.reportDTCFaultDetectionCounter,
            ReadDTCInformation.Subfunction.reportDTCWithPermanentStatus,

            # Documentation is confusing about reportDTCSnapshotIdentification subfunction.
            # It is presented with reportDTCSnapshotRecordByDTCNumber (2 params) but a footnote says that these 2 parameters
            # are not to be provided for reportDTCSnapshotIdentification. Therefore, it is the same as other no-params subfn
            ReadDTCInformation.Subfunction.reportDTCSnapshotIdentification

        ]

        request_subfn_status_mask = [
            ReadDTCInformation.Subfunction.reportNumberOfDTCByStatusMask,
            ReadDTCInformation.Subfunction.reportDTCByStatusMask,
            ReadDTCInformation.Subfunction.reportMirrorMemoryDTCByStatusMask,
            ReadDTCInformation.Subfunction.reportNumberOfMirrorMemoryDTCByStatusMask,
            ReadDTCInformation.Subfunction.reportNumberOfEmissionsRelatedOBDDTCByStatusMask,
            ReadDTCInformation.Subfunction.reportEmissionsRelatedOBDDTCByStatusMask
        ]

        request_subfn_mask_record_plus_snapshot_record_number = [
            ReadDTCInformation.Subfunction.reportDTCSnapshotRecordByDTCNumber
        ]

        request_subfn_mask_record_plus_snapshot_record_number_plus_memory_selection = [
            ReadDTCInformation.Subfunction.reportUserDefMemoryDTCSnapshotRecordByDTCNumber
        ]

        request_subfn_snapshot_record_number = [
            ReadDTCInformation.Subfunction.reportDTCSnapshotRecordByRecordNumber
        ]

        request_subfn_mask_record_plus_extdata_record_number = [
            ReadDTCInformation.Subfunction.reportDTCExtendedDataRecordByDTCNumber,
            ReadDTCInformation.Subfunction.reportMirrorMemoryDTCExtendedDataRecordByDTCNumber
        ]

        request_subfn_mask_record_plus_extdata_record_number_plus_memory_selection = [
            ReadDTCInformation.Subfunction.reportUserDefMemoryDTCExtDataRecordByDTCNumber,
        ]

        request_subfn_severity_plus_status_mask = [
            ReadDTCInformation.Subfunction.reportNumberOfDTCBySeverityMaskRecord,
            ReadDTCInformation.Subfunction.reportDTCBySeverityMaskRecord
        ]

        request_subfn_mask_record = [
            ReadDTCInformation.Subfunction.reportSeverityInformationOfDTC
        ]

        request_subfn_status_mask_plus_memory_selection = [
            ReadDTCInformation.Subfunction.reportUserDefMemoryDTCByStatusMask
        ]

        cls.check_subfunction_valid(subfunction, standard_version)

        if status_mask is not None and isinstance(status_mask, Dtc.Status):
            status_mask = status_mask.get_byte_as_int()

        if severity_mask is not None and isinstance(severity_mask, Dtc.Severity):
            severity_mask = severity_mask.get_byte_as_int()
            severity_mask = severity_mask & 0xE0

        if dtc_class is not None :
            if severity_mask is None:
                raise ValueError("A severity mask must be specified in order to specify a dataclass")
            
            if isinstance(dtc_class, Dtc.DtcClass):
                dtc_class = dtc_class.get_byte_as_int()
            severity_mask |= (dtc_class & 0x1F)

        if dtc is not None and isinstance(dtc, Dtc):
            dtc = dtc.id

        req = Request(service=cls, subfunction=subfunction)

        if subfunction in request_subfn_no_param:		# Service ID + Subfunction
            pass

        elif subfunction in request_subfn_status_mask:
            cls.assert_status_mask(status_mask, subfunction)
            req.data = struct.pack('B', status_mask)

        elif subfunction in request_subfn_mask_record_plus_snapshot_record_number:
            cls.assert_dtc(dtc, subfunction)
            assert dtc is not None  # mypy nitpick
            cls.assert_snapshot_record_number(snapshot_record_number, subfunction)
            req.data = cls.pack_dtc(dtc) + struct.pack('B', snapshot_record_number)

        elif subfunction in request_subfn_mask_record_plus_snapshot_record_number_plus_memory_selection:
            cls.assert_dtc(dtc, subfunction)
            cls.assert_snapshot_record_number(snapshot_record_number, subfunction)
            cls.assert_memory_selection(memory_selection, subfunction)
            assert dtc is not None  # mypy nitpick
            req.data = cls.pack_dtc(dtc) + struct.pack('BB', snapshot_record_number, memory_selection)

        elif subfunction in request_subfn_snapshot_record_number:
            cls.assert_snapshot_record_number(snapshot_record_number, subfunction)
            req.data = struct.pack('B', snapshot_record_number)

        elif subfunction in request_subfn_mask_record_plus_extdata_record_number:
            cls.assert_dtc(dtc, subfunction)
            assert dtc is not None  # mypy nitpick
            cls.assert_extended_data_record_number(extended_data_record_number, subfunction)
            req.data = cls.pack_dtc(dtc) + struct.pack('B', extended_data_record_number)

        elif subfunction in request_subfn_mask_record_plus_extdata_record_number_plus_memory_selection:
            cls.assert_dtc(dtc, subfunction)
            assert dtc is not None  # mypy nitpick
            cls.assert_memory_selection(memory_selection, subfunction)
            cls.assert_extended_data_record_number(extended_data_record_number, subfunction)
            req.data = cls.pack_dtc(dtc) + struct.pack('BB', extended_data_record_number, memory_selection)

        elif subfunction in request_subfn_severity_plus_status_mask:
            cls.assert_status_mask(status_mask, subfunction)
            cls.assert_severity_mask(severity_mask, subfunction)
            req.data = struct.pack('BB', severity_mask, status_mask)

        elif subfunction in request_subfn_mask_record:
            cls.assert_dtc(dtc, subfunction)
            assert dtc is not None
            req.data = cls.pack_dtc(dtc)

        elif subfunction in request_subfn_status_mask_plus_memory_selection:
            cls.assert_memory_selection(memory_selection, subfunction)
            cls.assert_status_mask(status_mask, subfunction)
            req.data = struct.pack('BB', status_mask, memory_selection)

        elif subfunction == ReadDTCInformation.Subfunction.reportDTCExtDataRecordByRecordNumber:
            cls.assert_extended_data_record_number(extended_data_record_number, subfunction, maxval=0xEF)  # Maximum specified by ISO-14229:2020
            req.data = struct.pack('B', extended_data_record_number)

        elif subfunction == ReadDTCInformation.Subfunction.reportWWHOBDDTCByMaskRecord:
            cls.assert_status_mask(status_mask, subfunction)
            cls.assert_severity_mask(severity_mask, subfunction)
            cls.assert_functional_group_id(functional_group_id, subfunction)  # Maximum specified by ISO-14229:2020 
            req.data = struct.pack('BBB', functional_group_id, status_mask, severity_mask)
       
        elif subfunction == ReadDTCInformation.Subfunction.reportWWHOBDDTCWithPermanentStatus:
            cls.assert_functional_group_id(functional_group_id, subfunction)  # Maximum specified by ISO-14229:2020 
            req.data = struct.pack('B', functional_group_id)

        return req

    @classmethod
    def interpret_response(cls,
                           response: Response,
                           subfunction: int,
                           extended_data_size: Optional[Union[int, Dict[int, int]]] = None,
                           tolerate_zero_padding: bool = True,
                           ignore_all_zero_dtc: bool = True,
                           dtc_snapshot_did_size: int = 2,
                           didconfig: Optional[DIDConfig] = None,
                           standard_version: int = latest_standard) -> InterpretedResponse:
        """
        Populates the response ``service_data`` property with an instance of :class:`ReadDTCInformation.ResponseData<udsoncan.services.ReadDTCInformation.ResponseData>`

        :param response: The received response to interpret
        :type response: :ref:`Response<Response>`

        :param subfunction: The service subfunction. Values are defined in :class:`ReadDTCInformation.Subfunction<udsoncan.services.ReadDTCInformation.Subfunction>`
        :type subfunction: int

        :param extended_data_size: Extended data size to expect. Extended data is implementation specific, therefore, size is not standardized
        :type extended_data_size: int or dict[int, int] to specify a size per extended data ID

        :param tolerate_zero_padding: Ignore trailing zeros in the response data avoiding raising false :class:`InvalidResponseException<udsoncan.exceptions.InvalidResponseException>`.
        :type tolerate_zero_padding:  bool

        :param ignore_all_zero_dtc: Discard any DTC entries that have an ID of 0. Avoid reading extra DTCs when using a transport protocol using zero padding.
        :type ignore_all_zero_dtc: bool

        :param dtc_snapshot_did_size: Number of bytes to encode the data identifier number. Other services such as :ref:`ReadDataByIdentifier<ReadDataByIdentifier>` encode DID over 2 bytes.
                UDS standard does not define the size of the snapshot DID, therefore, it must be supplied.
        :type dtc_snapshot_did_size: int

        :param didconfig: Definition of DID codecs. Dictionary mapping a DID (int) to a valid :ref:`DidCodec<DidCodec>` class or pack/unpack string 
        :type didconfig: dict[int] = :ref:`DidCodec<DidCodec>`

        :raises InvalidResponseException: If response length is wrong or does not match DID configuration
        :raises ValueError: If parameters are out of range, missing or wrong types
        :raises ConfigError: If the server returns a snapshot DID not defined in ``didconfig``
        """

        cls.check_subfunction_valid(subfunction, standard_version)

        # Response grouping for responses that are encoded the same way
        response_subfn_dtc_availability_mask_plus_dtc_record = [
            ReadDTCInformation.Subfunction.reportDTCByStatusMask,
            ReadDTCInformation.Subfunction.reportUserDefMemoryDTCByStatusMask,
            ReadDTCInformation.Subfunction.reportSupportedDTCs,
            ReadDTCInformation.Subfunction.reportFirstTestFailedDTC,
            ReadDTCInformation.Subfunction.reportFirstConfirmedDTC,
            ReadDTCInformation.Subfunction.reportMostRecentTestFailedDTC,
            ReadDTCInformation.Subfunction.reportMostRecentConfirmedDTC,
            ReadDTCInformation.Subfunction.reportMirrorMemoryDTCByStatusMask,
            ReadDTCInformation.Subfunction.reportEmissionsRelatedOBDDTCByStatusMask,
            ReadDTCInformation.Subfunction.reportDTCWithPermanentStatus
        ]

        response_subfn_number_of_dtc = [
            ReadDTCInformation.Subfunction.reportNumberOfDTCByStatusMask,
            ReadDTCInformation.Subfunction.reportNumberOfDTCBySeverityMaskRecord,
            ReadDTCInformation.Subfunction.reportNumberOfMirrorMemoryDTCByStatusMask,
            ReadDTCInformation.Subfunction.reportNumberOfEmissionsRelatedOBDDTCByStatusMask,
        ]

        response_subfn_dtc_availability_mask_plus_dtc_record_with_severity = [
            ReadDTCInformation.Subfunction.reportDTCBySeverityMaskRecord,
            ReadDTCInformation.Subfunction.reportSeverityInformationOfDTC
        ]

        response_subfn_dtc_plus_fault_counter = [
            ReadDTCInformation.Subfunction.reportDTCFaultDetectionCounter
        ]

        response_subfn_dtc_plus_sapshot_record = [
            ReadDTCInformation.Subfunction.reportDTCSnapshotIdentification
        ]

        response_sbfn_dtc_status_snapshots_records = [
            ReadDTCInformation.Subfunction.reportDTCSnapshotRecordByDTCNumber,
            ReadDTCInformation.Subfunction.reportUserDefMemoryDTCSnapshotRecordByDTCNumber
        ]

        response_sbfn_dtc_status_snapshots_records_record_first = [
            ReadDTCInformation.Subfunction.reportDTCSnapshotRecordByRecordNumber
        ]

        response_subfn_mask_record_plus_extdata = [
            ReadDTCInformation.Subfunction.reportDTCExtendedDataRecordByDTCNumber,
            ReadDTCInformation.Subfunction.reportMirrorMemoryDTCExtendedDataRecordByDTCNumber,
            ReadDTCInformation.Subfunction.reportUserDefMemoryDTCExtDataRecordByDTCNumber
        ]

        response_subfn_record_number_plus_dtc_mask_plus_extdata = [
            ReadDTCInformation.Subfunction.reportDTCExtDataRecordByRecordNumber
        ]

        subfunctions_with_memory_selection = [
            ReadDTCInformation.Subfunction.reportUserDefMemoryDTCByStatusMask,
            ReadDTCInformation.Subfunction.reportUserDefMemoryDTCSnapshotRecordByDTCNumber,
            ReadDTCInformation.Subfunction.reportUserDefMemoryDTCExtDataRecordByDTCNumber
        ]

        firstbyte_is_memory_selection_echo = (subfunction in subfunctions_with_memory_selection)

        if response.data is None:
            raise InvalidResponseException(response, "No data in response")

        if len(response.data) < 1:
            raise InvalidResponseException(response, 'Response must be at least 1 byte long (echo of subfunction)')

        response.service_data = cls.ResponseData(
            subfunction_echo=response.data[0],  # First byte is subfunction
        )

        # Now for each response group, we have a different decoding algorithm
        if subfunction in response_subfn_dtc_availability_mask_plus_dtc_record + response_subfn_dtc_availability_mask_plus_dtc_record_with_severity:

            if subfunction in response_subfn_dtc_availability_mask_plus_dtc_record:
                dtc_size = 4  # DTC ID (3) + Status (1)
            elif subfunction in response_subfn_dtc_availability_mask_plus_dtc_record_with_severity:
                dtc_size = 6  # DTC ID (3) + Status (1) + Severity (1) + FunctionalUnit (1)

            if firstbyte_is_memory_selection_echo:
                if len(response.data) < 3:
                    raise InvalidResponseException(
                        response, 'Response must be at least 3 bytes long (echo of subfunction, echo of MemorySelection and DTCStatusAvailabilityMask)')
                response.service_data.memory_selection_echo = int(response.data[1])
                actual_byte = 2
            else:
                if len(response.data) < 2:
                    raise InvalidResponseException(
                        response, 'Response must be at least 2 byte long (echo of subfunction and DTCStatusAvailabilityMask)')
                actual_byte = 1

            response.service_data.status_availability = Dtc.Status.from_byte(response.data[actual_byte])
            actual_byte += 1

            while True:  # Loop until we have read all dtcs
                if len(response.data) <= actual_byte:
                    break  # done

                elif len(response.data) < actual_byte + dtc_size:
                    partial_dtc_length = len(response.data) - actual_byte
                    if tolerate_zero_padding and response.data[actual_byte:] == b'\x00' * partial_dtc_length:
                        break
                    else:
                        # We purposely ignore extra byte for subfunction reportSeverityInformationOfDTC as it is supposed to return 0 or 1 DTC.
                        if subfunction != ReadDTCInformation.Subfunction.reportSeverityInformationOfDTC or actual_byte == 2:
                            raise InvalidResponseException(
                                response, 'Incomplete DTC record. Missing %d bytes to response to complete the record' % (dtc_size - partial_dtc_length))

                else:
                    dtc_bytes = response.data[actual_byte:actual_byte + dtc_size]
                    if dtc_bytes == b'\x00' * dtc_size and ignore_all_zero_dtc:
                        pass  # ignore
                    else:
                        if subfunction in response_subfn_dtc_availability_mask_plus_dtc_record:
                            dtc = Dtc(struct.unpack('>L', b'\x00' + dtc_bytes[0:3])[0])
                            dtc.status.set_byte(dtc_bytes[3])
                        elif subfunction in response_subfn_dtc_availability_mask_plus_dtc_record_with_severity:
                            dtc = Dtc(struct.unpack('>L', b'\x00' + dtc_bytes[2:5])[0])
                            dtc.severity.set_byte(dtc_bytes[0])
                            dtc.functional_unit = dtc_bytes[1]
                            dtc.status.set_byte(dtc_bytes[5])

                        response.service_data.dtcs.append(dtc)
                actual_byte += dtc_size
            response.service_data.dtc_count = len(response.service_data.dtcs)

        # The 2 following subfunction responses have different purposes but their constructions are very similar.
        elif subfunction in response_subfn_dtc_plus_fault_counter + response_subfn_dtc_plus_sapshot_record:
            dtc_size = 4
            if len(response.data) < 1:
                raise InvalidResponseException(response, 'Response must be at least 1 byte long (echo of subfunction)')

            actual_byte = 1 	# Increasing index
            dtc_map: Dict[int, Dtc] = dict()  # This map is used to append snapshot to existing DTC.

            while True: 	# Loop until we have read all dtcs
                if len(response.data) <= actual_byte:
                    break 	# done

                elif len(response.data) < actual_byte + dtc_size:
                    partial_dtc_length = len(response.data) - actual_byte
                    if tolerate_zero_padding and response.data[actual_byte:] == b'\x00' * partial_dtc_length:
                        break
                    else:
                        raise InvalidResponseException(
                            response, 'Incomplete DTC record. Missing %d bytes to response to complete the record' % (dtc_size - partial_dtc_length))
                else:
                    dtc_bytes = response.data[actual_byte:actual_byte + dtc_size]
                    if dtc_bytes == b'\x00' * dtc_size and ignore_all_zero_dtc:
                        pass  # ignore
                    else:
                        dtcid = struct.unpack('>L', b'\x00' + dtc_bytes[0:3])[0]
                        # We create the DTC or get its reference if already created.
                        dtc_created = False
                        if dtcid in dtc_map and subfunction in response_subfn_dtc_plus_sapshot_record:
                            dtc = dtc_map[dtcid]
                        else:
                            dtc = Dtc(dtcid)
                            dtc_map[dtcid] = dtc
                            dtc_created = True

                        # We either read the DTC fault counter or Snapshot record number.
                        if subfunction in response_subfn_dtc_plus_fault_counter:
                            dtc.fault_counter = dtc_bytes[3]

                        elif subfunction in response_subfn_dtc_plus_sapshot_record:
                            record_number = dtc_bytes[3]

                            if dtc.snapshots is None:
                                dtc.snapshots = []

                            dtc.snapshots.append(record_number)

                        # Adds the DTC to the list.
                        if dtc_created:
                            response.service_data.dtcs.append(dtc)

                actual_byte += dtc_size

            response.service_data.dtc_count = len(response.service_data.dtcs)

        # This group of responses returns a number of DTCs available
        elif subfunction in response_subfn_number_of_dtc:
            if len(response.data) < 5:
                raise InvalidResponseException(response, 'Response must be exactly 5 bytes long ')

            response.service_data.status_availability = Dtc.Status.from_byte(response.data[1])
            response.service_data.dtc_format = response.data[2]
            response.service_data.dtc_count = struct.unpack('>H', response.data[3:5])[0]

        # This group of responses returns DTC snapshots
        # Responses include a DTC, many snapshot records. For each record, we find many Data Identifiers.
        # We create one Dtc.Snapshot for each DID. That'll be easier to work with.
        # <DTC,RecordNumber1,NumberOfDid_X,DID1,DID2,...DIDX, RecordNumber2,NumberOfDid_Y,DID1,DID2,...DIDY, etc>
        elif subfunction in response_sbfn_dtc_status_snapshots_records:

            minlength = 5 if not firstbyte_is_memory_selection_echo else 6
            if len(response.data) < minlength:
                raise InvalidResponseException(response, 'Response must be at least %d bytes long' % minlength)

            if firstbyte_is_memory_selection_echo:
                response.service_data.memory_selection_echo = response.data[1]
                actual_byte = 2
            else:
                actual_byte = 1

            dtc = Dtc(struct.unpack('>L', b'\x00' + response.data[actual_byte:(actual_byte + 3)])[0])
            dtc.status.set_byte(response.data[actual_byte + 3])
            actual_byte += 4		# Increasing index

            tools.validate_int(dtc_snapshot_did_size, min=1, max=8, name='dtc_snapshot_did_size')

            while True:		# Loop until we have read all dtcs
                if len(response.data) <= actual_byte:
                    break  # done

                remaining_data = response.data[actual_byte:]
                if tolerate_zero_padding and remaining_data == b'\x00' * len(remaining_data):
                    break

                if len(remaining_data) < 2:
                    raise InvalidResponseException(response, 'Incomplete response from server. Missing "number of identifier" and following data')

                record_number = remaining_data[0]
                number_of_did = remaining_data[1]
                # Validate record number and number of DID before continuing
                if number_of_did == 0:
                    raise InvalidResponseException(response, 'Server returned snapshot record #%d with no data identifier included' % (record_number))

                if len(remaining_data) < 2 + dtc_snapshot_did_size:
                    raise InvalidResponseException(response, 'Incomplete response from server. Missing DID number and associated data.')

                actual_byte += 2
                for i in range(number_of_did):
                    remaining_data = response.data[actual_byte:]
                    snapshot = Dtc.Snapshot()  # One snapshot per DID for convenience.
                    snapshot.record_number = record_number

                    # As standard does not specify the length of the DID, we craft it based on a config
                    did = 0
                    for j in range(dtc_snapshot_did_size):
                        offset = dtc_snapshot_did_size - 1 - j
                        did |= (remaining_data[offset] << (8 * j))

                    # Decode the data based on DID number.
                    snapshot.did = did
                    didconfig = check_did_config(did, didconfig)
                    codec_definition = fetch_codec_definition_from_config(did, didconfig)
                    codec = make_did_codec_from_definition(codec_definition)

                    data_offset = dtc_snapshot_did_size
                    if len(remaining_data[data_offset:]) < len(codec):
                        raise InvalidResponseException(response, 'Incomplete response. Data for DID 0x%04x is only %d bytes while %d bytes is expected' % (
                            did, len(remaining_data[data_offset:]), len(codec)))

                    snapshot.raw_data = remaining_data[data_offset:data_offset + len(codec)]
                    snapshot.data = codec.decode(snapshot.raw_data)

                    dtc.snapshots.append(snapshot)
                    actual_byte += dtc_snapshot_did_size + len(codec)

            response.service_data.dtcs.append(dtc)
            response.service_data.dtc_count = 1

        # This group of responses returns DTC snapshots
        # Responses include a DTC, many snapshots records. For each record, we find many Data Identifiers.
        # We create one Dtc.Snapshot for each DID. That'll be easier to work with.
        # Similar to previous subfunction group, but order of information is changed.

        # <RecordNumber1, DTC1,NumberOfDid_X,DID1,DID2,...DIDX, RecordNumber2,DTC2, NumberOfDid_Y,DID1,DID2,...DIDY, etc>
        elif subfunction in response_sbfn_dtc_status_snapshots_records_record_first:
            tools.validate_int(dtc_snapshot_did_size, min=1, max=8, name='dtc_snapshot_did_size')

            if len(response.data) < 2:
                raise InvalidResponseException(response, 'Response must be at least 2 bytes long. Subfunction echo + RecordNumber ')

            actual_byte = 1	 # Increasing index
            while True:  # Loop through response data
                if len(response.data) <= actual_byte:
                    break  # done

                remaining_data = response.data[actual_byte:]
                record_number = remaining_data[0]

                # If empty response but filled with 0, it is considered ok
                if remaining_data == b'\x00' * len(remaining_data) and tolerate_zero_padding:
                    break

                # If record number received but no DTC provided (allowed according to standard), we exit.
                if len(remaining_data) == 1 or tolerate_zero_padding and remaining_data[1:] == b'\x00' * len(remaining_data[1:]):
                    break

                if len(remaining_data) < 5: 	# Partial DTC (No DTC at all is checked above)
                    raise InvalidResponseException(response, 'Incomplete response from server. Missing "DTCAndStatusRecord" and following data')

                if len(remaining_data) < 6:
                    raise InvalidResponseException(response, 'Incomplete response from server. Missing number of data identifier')

                # DTC decoding
                dtc = Dtc(struct.unpack('>L', b'\x00' + remaining_data[1:4])[0])
                dtc.status.set_byte(remaining_data[4])
                number_of_did = remaining_data[5]

                actual_byte += 6
                remaining_data = response.data[actual_byte:]

                if number_of_did == 0:
                    raise InvalidResponseException(response, 'Server returned snapshot record #%d with no data identifier included' % (record_number))

                if len(remaining_data) < dtc_snapshot_did_size:
                    raise InvalidResponseException(response, 'Incomplete response from server. Missing DID and associated data')

                # We have a DTC and 0 DID, next loop
                if tolerate_zero_padding and remaining_data == b'\x00' * len(remaining_data):
                    break

                # For each DID
                for i in range(number_of_did):
                    remaining_data = response.data[actual_byte:]
                    snapshot = Dtc.Snapshot()  # One snapshot epr DID for convenience
                    snapshot.record_number = record_number

                    # As standard does not specify the length of the DID, we craft it based on a config
                    did = 0
                    for j in range(dtc_snapshot_did_size):
                        offset = dtc_snapshot_did_size - 1 - j
                        did |= (remaining_data[offset] << (8 * j))

                    # Decode the data based on DID number.
                    snapshot.did = did
                    didconfig = check_did_config(did, didconfig)
                    codec_definition = fetch_codec_definition_from_config(did, didconfig)
                    codec = make_did_codec_from_definition(codec_definition)

                    data_offset = dtc_snapshot_did_size
                    if len(remaining_data[data_offset:]) < len(codec):
                        raise InvalidResponseException(response, 'Incomplete response. Data for DID 0x%04x is only %d bytes while %d bytes is expected' % (
                            did, len(remaining_data[data_offset:]), len(codec)))

                    snapshot.raw_data = remaining_data[data_offset:data_offset + len(codec)]
                    snapshot.data = codec.decode(snapshot.raw_data)

                    dtc.snapshots.append(snapshot)

                    actual_byte += dtc_snapshot_did_size + len(codec)

                response.service_data.dtcs.append(dtc)
            response.service_data.dtc_count = len(response.service_data.dtcs)

        # These subfunctions include DTC ExtraData. We give it raw to user.
        elif subfunction in response_subfn_mask_record_plus_extdata:
            cls.assert_extended_data_size_int_or_dict(extended_data_size, subfunction)
            assert extended_data_size is not None

            minlength = 5 if not firstbyte_is_memory_selection_echo else 6

            if len(response.data) < minlength:
                missing_data = 'MemorySelection and DTCAndStatusRecord' if firstbyte_is_memory_selection_echo else 'DTCAndStatusRecord'
                raise InvalidResponseException(response, 'Incomplete response from server. Missing %s' % missing_data)

            if firstbyte_is_memory_selection_echo:
                response.service_data.memory_selection_echo = response.data[1]
                actual_byte = 2
            else:
                actual_byte = 1

            # DTC decoding
            dtc = Dtc(struct.unpack('>L', b'\x00' + response.data[actual_byte:(actual_byte + 3)])[0])
            dtc.status.set_byte(response.data[actual_byte + 3])

            size = cls.get_extended_data_size(dtc, extended_data_size)

            actual_byte = actual_byte + 4  # Increasing index
            while actual_byte < len(response.data):  # Loop through data
                remaining_data = response.data[actual_byte:]
                record_number = remaining_data[0]

                if record_number == 0:
                    if remaining_data == b'\x00' * len(remaining_data) and tolerate_zero_padding:
                        break
                    else:
                        raise InvalidResponseException(
                            response, 'Extended data record number given by the server is 0 but this value is a reserved value.')

                actual_byte += 1
                remaining_data = response.data[actual_byte:]

                if len(remaining_data) < size:
                    raise InvalidResponseException(response, 'Incomplete response from server. Length of extended data for DTC 0x%06x with record number 0x%02x is %d bytes but smaller than given data_size of %d bytes' % (
                        dtc.id, record_number, len(remaining_data), size))

                exdata = Dtc.ExtendedData()
                exdata.record_number = record_number
                exdata.raw_data = remaining_data[0:size]

                dtc.extended_data.append(exdata)

                actual_byte += size

            response.service_data.dtcs.append(dtc)
            response.service_data.dtc_count = len(response.service_data.dtcs)

        elif subfunction in response_subfn_record_number_plus_dtc_mask_plus_extdata:
            cls.assert_extended_data_size_int_or_dict(extended_data_size, subfunction)
            assert extended_data_size is not None

            if len(response.data) < 2:
                raise InvalidResponseException(response, 'Incomplete response from server. Missing DTCExtDataRecordNumber')

            record_number = int(response.data[1])
            if record_number > 0xEF:
                raise InvalidResponseException(
                    response, 'Server returned a RecordNumber of %d which is out of range (00-EF) according to ISO-14229:2020', record_number)

            actual_byte = 2
            received_dtcs = set()
            while True:
                if actual_byte == len(response.data):
                    break

                remaining_data = response.data[actual_byte:]
                all_zero_ending = True if remaining_data == b'\x00' * len(remaining_data) else False

                if all_zero_ending:
                    try:
                        zero_dtc_size = cls.get_extended_data_size(Dtc(0), extended_data_size)
                    except:
                        zero_dtc_size = None

                    all_zero_dtc_to_read = (zero_dtc_size is not None and len(remaining_data) >= zero_dtc_size + 4) and ~ignore_all_zero_dtc

                    if ~all_zero_dtc_to_read:
                        if tolerate_zero_padding:
                            break
                        else:
                            raise InvalidResponseException(response, 'Server response ends with a sequence of zero and zero adding is not tolerated')

                # DTC decoding
                if len(response.data) - actual_byte < 4:
                    raise InvalidResponseException(response, 'Incomplete DTCAndStatusRecord at position %d' % actual_byte)

                dtc = Dtc(struct.unpack('>L', b'\x00' + response.data[actual_byte:(actual_byte + 3)])[0])
                if dtc.id in received_dtcs:
                    raise InvalidResponseException(
                        response, 'The server returned two set of data for the same DTC (0x%06X) and same record number (%d)', (dtc.id, record_number))

                received_dtcs.add(dtc.id)

                dtc.status.set_byte(response.data[actual_byte + 3])
                actual_byte += 4

                size = cls.get_extended_data_size(dtc, extended_data_size)

                if len(response.data) - actual_byte < size:
                    raise InvalidResponseException(response, 'Incomplete ExtendedData for DTC 0x%06X. Configuration says that there should be %d bytes of data but server only gave %d bytes.' % (
                        dtc.id, size, len(response.data) - actual_byte))

                exdata = Dtc.ExtendedData()
                exdata.record_number = record_number
                exdata.raw_data = response.data[actual_byte:actual_byte + size]
                dtc.extended_data.append(exdata)
                response.service_data.dtcs.append(dtc)

                actual_byte += size

            response.service_data.dtc_count = len(response.service_data.dtcs)

        elif subfunction in [ReadDTCInformation.Subfunction.reportWWHOBDDTCByMaskRecord, ReadDTCInformation.Subfunction.reportWWHOBDDTCWithPermanentStatus]:

            if subfunction == ReadDTCInformation.Subfunction.reportWWHOBDDTCByMaskRecord:
                if len(response.data) < 5:
                    raise InvalidResponseException(response, 'Incomplete response from server.')
                
                response.service_data.functional_group_id = response.data[1]
                response.service_data.status_availability = Dtc.Status.from_byte(response.data[2])
                response.service_data.severity_availability = Dtc.Severity.from_byte(response.data[3])
                response.service_data.dtc_format = response.data[4]
                remaining_bytes = response.data[5:]
            elif subfunction == ReadDTCInformation.Subfunction.reportWWHOBDDTCWithPermanentStatus:
                if len(response.data) < 4:
                    raise InvalidResponseException(response, 'Incomplete response from server.')
                
                response.service_data.functional_group_id = response.data[1]
                response.service_data.status_availability = Dtc.Status.from_byte(response.data[2])
                response.service_data.dtc_format = response.data[3]
                remaining_bytes = response.data[4:]
            else:
                raise NotImplementedError("Unreachable code")

            if response.service_data.functional_group_id > 0xFE:
                raise InvalidResponseException(response, "FunctionalGroupIdentifier returned by the server is not smaller or equal than 0xFE")
            
            if response.service_data.dtc_format not in [Dtc.Format.SAE_J2012_DA_DTCFormat_04, Dtc.Format.SAE_J1939_73]:
                raise InvalidResponseException(response, "DTCFormatIdentifier returned by the server is not one of the following: SAE_J2012-DA_DTCFormat_04 (4), SAE_J1939-73_DTCFormat(2). Got 0x%02x" % response.service_data.dtc_format)

            if len(remaining_bytes) % 5 != 0:
                raise InvalidResponseException(response, 'Incomplete response from server. Remaining bytes must be a multiple of 5')

            while remaining_bytes:
                severity = remaining_bytes[0]
                dtc = Dtc(struct.unpack('>L', b'\x00' + remaining_bytes[1:4])[0])
                status_of_dtc = Dtc.Status.from_byte(remaining_bytes[4])
                dtc.severity.set_byte(severity)
                dtc.status = status_of_dtc
                remaining_bytes = remaining_bytes[5:]
                response.service_data.dtcs.append(dtc)

            response.service_data.dtc_count = len(response.service_data.dtcs)

        return cast(ReadDTCInformation.InterpretedResponse, response)

    @classmethod
    def get_extended_data_size(cls, dtc: Dtc, given_size: Union[int, Dict[int, int]]) -> int:
        size = None
        if isinstance(given_size, int):
            size = given_size
        elif isinstance(given_size, dict):
            if dtc.id not in given_size:
                raise ConfigError(key=dtc.id, msg='Server returned extended data for DTC 0x%06X but no data size was specified for this DTC' % (dtc.id))
            size = given_size[dtc.id]
        else:
            raise ValueError('Unsupported extended data size type.')

        return size



from udsoncan.BaseService import BaseService, BaseResponseData
from udsoncan.ResponseCode import ResponseCode
from udsoncan.exceptions import *
from udsoncan.Response import Response
from udsoncan.Request import Request

from typing import cast


class TesterPresent(BaseService):
    _sid = 0x3E

    supported_negative_response = [ResponseCode.SubFunctionNotSupported,
                                   ResponseCode.IncorrectMessageLengthOrInvalidFormat
                                   ]

    class ResponseData(BaseResponseData):
        """
        .. data:: subfunction_echo

                Requests subfunction echoed back by the server. This value should always be 0
        """

        subfunction_echo: int

        def __init__(self, subfunction_echo: int):
            super().__init__(TesterPresent)
            self.subfunction_echo = subfunction_echo

    class InterpretedResponse(Response):
        service_data: "TesterPresent.ResponseData"

    @classmethod
    def make_request(cls) -> Request:
        """
        Generates a request for TesterPresent
        """
        return Request(service=cls, subfunction=0)

    @classmethod
    def interpret_response(cls, response: Response) -> InterpretedResponse:
        """
        Populates the response ``service_data`` property with an instance of :class:`TesterPresent.ResponseData<udsoncan.services.TesterPresent.ResponseData>`

        :param response: The received response to interpret
        :type response: :ref:`Response<Response>`

        :raises InvalidResponseException: If length of ``response.data`` is too short
        """
        if response.data is None:
            raise InvalidResponseException(response, "No data in response")

        if len(response.data) < 1:
            raise InvalidResponseException(response, "Response data must be at least 1 bytes")

        response.service_data = cls.ResponseData(
            subfunction_echo=response.data[0]
        )

        return cast(TesterPresent.InterpretedResponse, response)



from udsoncan.typing import ClientConfig
from udsoncan import latest_standard
from typing import cast

default_client_config: ClientConfig = cast(ClientConfig, {
    'exception_on_negative_response': True,
    'exception_on_invalid_response': True,
    'exception_on_unexpected_response': True,
    'security_algo': None,
    'security_algo_params': None,
    'tolerate_zero_padding': True,
    'ignore_all_zero_dtc': True,
    'dtc_snapshot_did_size': 2,		# Not specified in standard. 2 bytes matches other services format.
    'server_address_format': None,		# 8,16,24,32,40
    'server_memorysize_format': None,		# 8,16,24,32,40
    'data_identifiers': {},
    'input_output': {},
    'request_timeout': 5,
    'p2_timeout': 1,
    'p2_star_timeout': 5,
    'standard_version': latest_standard,  # 2006, 2013, 2020
    'use_server_timing': True,
    'extended_data_size': None,
    'nrc78_callback':None
})



__all__ = ['Baudrate']

import struct


class Baudrate:
    """
    Represents a link speed in bit per seconds (or symbol per seconds to be more accurate).
    This class is used by the :ref:`LinkControl<LinkControl>` service that controls the underlying protocol speeds.

    The class can encode the baudrate in 2 different fashions : **Fixed** or **Specific**.

    Some standard baudrate values are defined within ISO-14229:2006 Annex B.3

    :param baudrate: The baudrate to be used. 
    :type baudrate: int

    :param baudtype: Tells how the baudrate shall be encoded. 4 values are possible:

            - ``Baudrate.Type.Fixed`` (0) : Will encode the baudrate in a single byte Fixed fashion. `baudrate` should be a supported value such as 9600, 19200, 125000, 250000, etc.
            - ``Baudrate.Type.Specific`` (1) : Will encode the baudrate in a three-byte Specific fashion. `baudrate` can be any value ranging from 0 to 0xFFFFFF
            - ``Baudrate.Type.Identifier`` (2) : Will encode the baudrate in a single byte Fixed fashion. `baudrate` should be the byte value to encode if the user wants to use a custom type.
            - ``Baudrate.Type.Auto`` (3) : Let the class guess the type. 

                    - If ``baudrate`` is a known standard value (19200, 38400, etc), then Fixed shall be used
                    - If ``baudrate`` is an integer that fits in a single byte, then Identifier shall be used
                    - If ``baudrate`` is none of the above, then Specific will be used.
    :type baudtype: int

    """
    baudrate_map = {
        9600: 0x01,
        19200: 0x02,
        38400: 0x03,
        57600: 0x04,
        115200: 0x05,
        125000: 0x10,
        250000: 0x11,
        500000: 0x12,
        1000000: 0x13,
    }

    class Type:
        Fixed = 0		# When baudrate is a predefined value from standard
        Specific = 1  # When using custom baudrate
        Identifier = 2  # Baudrate implied by baudrate ID
        Auto = 3		# Let the class decide the type

    baudtype: int
    baudrate: int

    # User can specify the type of baudrate or let this class guess what he wants (this adds some simplicity for non-experts).
    def __init__(self, baudrate: int, baudtype: int = Type.Auto):
        if not isinstance(baudrate, int):
            raise ValueError('baudrate must be an integer')

        if baudrate < 0:
            raise ValueError('baudrate must be an integer greater than 0')

        if baudtype == self.Type.Auto:
            if baudrate in self.baudrate_map:
                self.baudtype = self.Type.Fixed
            else:
                if baudrate <= 0xFF:
                    self.baudtype = self.Type.Identifier
                else:
                    self.baudtype = self.Type.Specific
        else:
            self.baudtype = baudtype

        if self.baudtype == self.Type.Specific:
            if baudrate > 0xFFFFFF:
                raise ValueError('Baudrate value cannot be bigger than a 24 bits value.')

        elif self.baudtype == self.Type.Identifier:
            if baudrate > 0xFF:
                raise ValueError('Baudrate ID must be an integer between 0 and 0xFF')
        elif self.baudtype == self.Type.Fixed:
            if baudrate not in self.baudrate_map:
                raise ValueError('baudrate must be part of the supported baudrate list defined by UDS standard')
        else:
            raise ValueError('Unknown baudtype : %s' % self.baudtype)

        self.baudrate = baudrate

    # internal helper to change the type of this baudrate
    def make_new_type(self, baudtype: int) -> "Baudrate":
        if baudtype not in [self.Type.Fixed, self.Type.Specific]:
            raise ValueError('Baudrate type can only be change to Fixed or Specific')

        return Baudrate(self.effective_baudrate(), baudtype=baudtype)

    # Returns the baudrate in Symbol Per Seconds if available, otherwise value given by the user.
    def effective_baudrate(self) -> int:
        if self.baudtype == self.Type.Identifier:
            for k in self.baudrate_map:
                if self.baudrate_map[k] == self.baudrate:
                    return k

            raise RuntimeError('Unknown effective baudrate, this could indicate a bug')
        else:
            return self.baudrate

    # Encodes the baudrate value the way they are exchanged.
    def get_bytes(self) -> bytes:
        if self.baudtype == self.Type.Fixed:
            return struct.pack('B', self.baudrate_map[self.baudrate])

        if self.baudtype == self.Type.Specific:
            b1 = (self.baudrate >> 16) & 0xFF
            b2 = (self.baudrate >> 8) & 0xFF
            b3 = (self.baudrate >> 0) & 0xFF
            return struct.pack('BBB', b1, b2, b3)

        if self.baudtype == self.Type.Identifier:
            return struct.pack('B', self.baudrate)

        raise RuntimeError('Unknown baudrate baudtype : %s' % self.baudtype)

    def __str__(self) -> str:
        baudtype_str = ''
        if self.baudtype == self.Type.Fixed:
            baudtype_str = 'Fixed'
        elif self.baudtype == self.Type.Specific:
            baudtype_str = 'Specific'
        elif self.baudtype == self.Type.Identifier:
            baudtype_str = 'Defined by identifier'

        return '%sBauds, %s format.' % (str(self.effective_baudrate()), baudtype_str)

    def __repr__(self):
        return '<%s: %s at 0x%08x>' % (self.__class__.__name__, str(self), id(self))



import struct
from udsoncan import DynamicDidDefinition
from udsoncan.Request import Request
from udsoncan.Response import Response
from udsoncan.exceptions import *
from udsoncan.BaseService import BaseService, BaseSubfunction, BaseResponseData
from udsoncan.ResponseCode import ResponseCode
import udsoncan.tools as tools
from typing import Optional, cast, List


class DynamicallyDefineDataIdentifier(BaseService):
    _sid = 0x2C

    supported_negative_response = [ResponseCode.SubFunctionNotSupported,
                                   ResponseCode.IncorrectMessageLengthOrInvalidFormat,
                                   ResponseCode.ConditionsNotCorrect,
                                   ResponseCode.RequestOutOfRange
                                   ]

    class Subfunction(BaseSubfunction):
        """
        DynamicallyDefineDataIdentifier defined subfunctions
        """
        __pretty_name__ = 'subfunction'

        defineByIdentifier = 1
        defineByMemoryAddress = 2
        clearDynamicallyDefinedDataIdentifier = 3

    class ResponseData(BaseResponseData):
        subfunction_echo: int
        did_echo: Optional[int]

        def __init__(self, subfunction_echo: int, did_echo: Optional[int] = None):
            super().__init__(DynamicallyDefineDataIdentifier)
            self.subfunction_echo = subfunction_echo
            self.did_echo = did_echo

    class InterpretedResponse(Response):
        service_data: "DynamicallyDefineDataIdentifier.ResponseData"

    @classmethod
    def make_request(cls, subfunction: int, did: Optional[int] = None, diddef: Optional[DynamicDidDefinition] = None) -> Request:
        """
        Generates a request for DynamicallyDefineDataIdentifier

        :param subfunction: Service subfunction. Allowed values are from 1 to 3
        :type subfunction: int

        :param did: The Data Identifier to define. Values from 0x0000 to 0xFFFF
        :type did: int

        :param diddef: Definition of the DID. Either by source DID or memory address. This parameter is only needed with subfunctions defineByIdentifie (1)` and defineByMemoryAddress (2)
        :type diddef: :ref:`DynamicDidDefinition<DynamicDidDefinition>`

        :raises ValueError: If parameters are out of range, missing or wrong type
        """
        tools.validate_int(subfunction, min=1, max=3, name='Subfunction')
        req = Request(service=cls, subfunction=subfunction)
        req.data = bytes()
        if subfunction in [cls.Subfunction.defineByIdentifier, cls.Subfunction.defineByMemoryAddress]:
            if not isinstance(diddef, DynamicDidDefinition):
                raise ValueError('A DynamicDidDefinition must be given to define a dynamic did with subfunction %d' % (subfunction))

            if did is None:
                raise ValueError('A DID number must be given with subfunction %d' % (subfunction))

        if did is not None:
            tools.validate_int(did, min=0, max=0xFFFF, name='DID number')

        if subfunction in [cls.Subfunction.defineByIdentifier, cls.Subfunction.defineByMemoryAddress]:
            if diddef is None:
                raise ValueError('DynamicDidDefinition must be given for this subfunction')

            diddef_entries = diddef.get()
            if len(diddef_entries) == 0:
                raise ValueError('DynamicDidDefinition object must have at least one DID specification')
            req.data = struct.pack('>H', did)

        if subfunction == cls.Subfunction.defineByIdentifier:
            assert diddef is not None
            if not diddef.is_by_source_did():
                raise ValueError("DynamicDidDefinition must be defined by source DID when used with subfunction 'defineByIdentifier'")
            diddef_entries = cast(List[DynamicDidDefinition.ByDidDefinition], diddef_entries)
            for entry_by_did in diddef_entries:
                req.data += struct.pack('>HBB', entry_by_did.source_did, entry_by_did.position, entry_by_did.memorysize)

        elif subfunction == cls.Subfunction.defineByMemoryAddress:
            assert diddef is not None
            if not diddef.is_by_memory_address():
                raise ValueError("DynamicDidDefinition must be defined by memory address when used with subfunction 'defineByMemoryAddress'")

            req.data += diddef.get_alfid().get_byte()
            diddef_entries = cast(List[DynamicDidDefinition.ByMemloc], diddef_entries)
            for entry_by_memloc in diddef_entries:
                req.data += entry_by_memloc.memloc.get_address_bytes()
                req.data += entry_by_memloc.memloc.get_memorysize_bytes()

        elif subfunction == cls.Subfunction.clearDynamicallyDefinedDataIdentifier:
            if did is not None:
                req.data = struct.pack('>H', did)

        return req

    @classmethod
    def interpret_response(cls, response: Response) -> InterpretedResponse:
        """
        Populates the response ``service_data`` property with an instance of :class:`DynamicallyDefineDataIdentifier.ResponseData<udsoncan.services.DynamicallyDefineDataIdentifier.ResponseData>`

        :param response: The received response to interpret
        :type response: :ref:`Response<Response>`

        :raises InvalidResponseException: If length of ``response.data`` is too short
        """
        if response.data is None:
            raise InvalidResponseException(response, "No data in response")

        if len(response.data) < 1:
            raise InvalidResponseException(response, "Response data must be at least 1 bytes")

        response.service_data = cls.ResponseData(
            subfunction_echo=int(response.data[0])
        )

        if response.service_data.subfunction_echo in [cls.Subfunction.defineByIdentifier, cls.Subfunction.defineByMemoryAddress]:
            if len(response.data) < 3:
                raise InvalidResponseException(response, "Missing or incomplete DID echo in response")

        if len(response.data) >= 3:
            response.service_data.did_echo = struct.unpack('>H', response.data[1:3])[0]

        return cast(DynamicallyDefineDataIdentifier.InterpretedResponse, response)



from udsoncan.Request import Request
from udsoncan.Response import Response
from udsoncan import CommunicationType
from udsoncan.exceptions import *
from udsoncan.BaseService import BaseService, BaseSubfunction, BaseResponseData
from udsoncan.ResponseCode import ResponseCode
import udsoncan.tools as tools
from udsoncan import latest_standard
import struct

from typing import cast, Union, Optional


class CommunicationControl(BaseService):
    _sid = 0x28

    class ControlType(BaseSubfunction):
        """
        CommunicationControl defined subfunctions
        """
        __pretty_name__ = 'control type'

        enableRxAndTx = 0
        enableRxAndDisableTx = 1
        disableRxAndEnableTx = 2
        disableRxAndTx = 3
        enableRxAndDisableTxWithEnhancedAddressInformation = 4
        enableRxAndTxWithEnhancedAddressInformation = 5

    supported_negative_response = [ResponseCode.SubFunctionNotSupported,
                                   ResponseCode.IncorrectMessageLengthOrInvalidFormat,
                                   ResponseCode.ConditionsNotCorrect,
                                   ResponseCode.RequestOutOfRange
                                   ]

    class ResponseData(BaseResponseData):
        """
        .. data:: control_type_echo

                Request subfunction echoed back by the server
        """
        control_type_echo: int

        def __init__(self, control_type_echo: int):
            super().__init__(CommunicationControl)
            self.control_type_echo = control_type_echo

    class InterpretedResponse(Response):
        service_data: "CommunicationControl.ResponseData"

    @classmethod
    def normalize_communication_type(self, communication_type: Union[int, bytes, CommunicationType]) -> CommunicationType:
        if not isinstance(communication_type, CommunicationType) and not isinstance(communication_type, int) and not isinstance(communication_type, bytes):
            raise ValueError('communication_type must either be a CommunicationType object or an integer')

        if isinstance(communication_type, int) or isinstance(communication_type, bytes):
            communication_type = CommunicationType.from_byte(communication_type)

        return communication_type

    @classmethod
    def make_request(cls, control_type: int, communication_type: CommunicationType, node_id: Optional[int] = None, standard_version=latest_standard) -> Request:
        """
        Generates a request for CommunicationControl

        :param control_type: Service subfunction. Allowed values are from 0 to 0x7F
        :type control_type: int

        :param communication_type: The communication type requested.
        :type communication_type: :ref:`CommunicationType <CommunicationType>`, int, bytes

        :param node_id: DTC memory identifier. This value is user defined and introduced in 2013 version of ISO-14229-1. 
            Possible and required only when ``control_type`` is ``enableRxAndDisableTxWithEnhancedAddressInformation`` or ``enableRxAndTxWithEnhancedAddressInformation``
            Default : ``None``
        :type node_id: int

        :param standard_version: The version of the ISO-14229 (the year). eg. 2006, 2013, 2020
        :type standard_version: int

        :raises ValueError: If parameters are out of range, missing or wrong type
        """
        tools.validate_int(control_type, min=0, max=0x7F, name='Control type')

        require_node_id = standard_version >= 2013 and control_type in (
            CommunicationControl.ControlType.enableRxAndDisableTxWithEnhancedAddressInformation,
            CommunicationControl.ControlType.enableRxAndTxWithEnhancedAddressInformation
        )

        if require_node_id and node_id is None:
            raise ValueError(
                "node_id is required when the standard version is 2013 (or more recent) and when control_type is enableRxAndDisableTxWithEnhancedAddressInformation or enableRxAndTxWithEnhancedAddressInformation ")
        elif not require_node_id and node_id is not None:
            raise ValueError(
                "node_id is only possible when the standard version is 2013 (or more recent) and when control_type is enableRxAndDisableTxWithEnhancedAddressInformation or enableRxAndTxWithEnhancedAddressInformation ")

        communication_type = cls.normalize_communication_type(communication_type)
        request = Request(service=cls, subfunction=control_type)
        payload = communication_type.get_byte()

        if node_id is not None:
            tools.validate_int(node_id, min=0, max=0xFFFF, name='nodeIdentificationNumber')
            payload += struct.pack('>H', node_id)

        request.data = payload
        return request

    @classmethod
    def interpret_response(cls, response: Response) -> InterpretedResponse:
        """
        Populates the response ``service_data`` property with an instance of :class:`CommunicationControl.ResponseData<udsoncan.services.CommunicationControl.ResponseData>`

        :param response: The received response to interpret
        :type response: :ref:`Response<Response>`

        :raises InvalidResponseException: If length of ``response.data`` is too short
        """
        if response.data is None:
            raise InvalidResponseException(response, "No data in response")

        if len(response.data) < 1:
            raise InvalidResponseException(response, "Response data must be at least 1 byte")

        response.service_data = cls.ResponseData(
            control_type_echo=response.data[0]
        )

        return cast(CommunicationControl.InterpretedResponse, response)



__all__ = ['Units']

from typing import Optional


class Units:
    # As defined in ISO-14229:2006 Annex C
    class Prefixs:
        class Prefix:
            id: int
            name: str
            symbol: str
            description: Optional[str]

            def __init__(self, id: int, name: str, symbol: str, description: Optional[str] = None):
                self.name = name
                self.id = id
                self.symbol = symbol
                self.description = description

            def __str__(self) -> str:
                return self.name

            def __repr__(self) -> str:
                desc = "(%s) " % self.description if self.description is not None else ""
                return "<UDS Unit prefix : %s[%s] %swith ID=%d at %08x>" % (self.name, self.symbol, desc, self.id, id(self))

        exa = Prefix(id=0x40, name='exa', symbol='E', description='10e18')
        peta = Prefix(id=0x41, name='peta', symbol='P', description='10e15')
        tera = Prefix(id=0x42, name='tera', symbol='T', description='10e12')
        giga = Prefix(id=0x43, name='giga', symbol='G', description='10e9')
        mega = Prefix(id=0x44, name='mega', symbol='M', description='10e6')
        kilo = Prefix(id=0x45, name='kilo', symbol='k', description='10e3')
        hecto = Prefix(id=0x46, name='hecto', symbol='h', description='10e2')
        deca = Prefix(id=0x47, name='deca', symbol='da', description='10e1')
        deci = Prefix(id=0x48, name='deci', symbol='d', description='10e-1')
        centi = Prefix(id=0x49, name='centi', symbol='c', description='10e-2')
        milli = Prefix(id=0x4A, name='milli', symbol='m', description='10e-3')
        micro = Prefix(id=0x4B, name='micro', symbol='m', description='10e-6')
        nano = Prefix(id=0x4C, name='nano', symbol='n', description='10e-9')
        pico = Prefix(id=0x4D, name='pico', symbol='p', description='10e-12')
        femto = Prefix(id=0x4E, name='femto', symbol='f', description='10e-15')
        atto = Prefix(id=0x4F, name='atto', symbol='a', description='10e-18')

    class Unit:
        id: int
        name: str
        symbol: str
        description: Optional[str]

        def __init__(self, id: int, name: str, symbol: str, description: Optional[str] = None):
            self.id = id
            self.name = name
            self.symbol = symbol
            self.description = description

        def __str__(self):
            return self.name

        def __repr__(self):
            desc = "(unit of %s) " % self.description if self.description is not None else ""
            return "<UDS Unit : %s[%s] %swith ID=%d at %08x>" % (self.name, self.symbol, desc, self.id, id(self))

    no_unit = Unit(id=0x00, name='no unit', symbol='-', description='-')
    meter = Unit(id=0x01, name='meter', symbol='m', description='length')
    foor = Unit(id=0x02, name='foot', symbol='ft', description='length')
    inch = Unit(id=0x03, name='inch', symbol='in', description='length')
    yard = Unit(id=0x04, name='yard', symbol='yd', description='length')
    english_mile = Unit(id=0x05, name='mile (English)', symbol='mi', description='length')
    gram = Unit(id=0x06, name='gram', symbol='g', description='mass')
    metric_ton = Unit(id=0x07, name='ton (metric)', symbol='t', description='mass')
    second = Unit(id=0x08, name='second', symbol='s', description='time')
    minute = Unit(id=0x09, name='minute', symbol='min', description='time')
    hour = Unit(id=0x0A, name='hour', symbol='h', description='time')
    day = Unit(id=0x0B, name='day', symbol='d', description='time')
    year = Unit(id=0x0C, name='year', symbol='y', description='time')
    ampere = Unit(id=0x0D, name='ampere', symbol='A', description='current')
    volt = Unit(id=0x0E, name='volt', symbol='V', description='voltage')
    coulomb = Unit(id=0x0F, name='coulomb', symbol='C', description='electric charge')
    ohm = Unit(id=0x10, name='ohm', symbol='W', description='resistance')
    farad = Unit(id=0x11, name='farad', symbol='F', description='capacitance')
    henry = Unit(id=0x12, name='henry', symbol='H', description='inductance')
    siemens = Unit(id=0x13, name='siemens', symbol='S', description='electric conductance')
    weber = Unit(id=0x14, name='weber', symbol='Wb', description='magnetic flux')
    tesla = Unit(id=0x15, name='tesla', symbol='T', description='magnetic flux density')
    kelvin = Unit(id=0x16, name='kelvin', symbol='K', description='thermodynamic temperature')
    Celsius = Unit(id=0x17, name='Celsius', symbol='°C', description='thermodynamic temperature')
    Fahrenheit = Unit(id=0x18, name='Fahrenheit', symbol='°F', description='thermodynamic temperature')
    candela = Unit(id=0x19, name='candela', symbol='cd', description='luminous intensity')
    radian = Unit(id=0x1A, name='radian', symbol='rad', description='plane angle')
    degree = Unit(id=0x1B, name='degree', symbol='°', description='plane angle')
    hertz = Unit(id=0x1C, name='hertz', symbol='Hz', description='frequency')
    joule = Unit(id=0x1D, name='joule', symbol='J', description='energy')
    Newton = Unit(id=0x1E, name='Newton', symbol='N', description='force')
    kilopond = Unit(id=0x1F, name='kilopond', symbol='kp', description='force')
    pound = Unit(id=0x20, name='pound force', symbol='lbf', description='force')
    watt = Unit(id=0x21, name='watt', symbol='W', description='power')
    horse = Unit(id=0x22, name='horse power (metric)', symbol='hk', description='power')
    horse = Unit(id=0x23, name='horse power(UK and US)', symbol='hp', description='power')
    Pascal = Unit(id=0x24, name='Pascal', symbol='Pa', description='pressure')
    bar = Unit(id=0x25, name='bar', symbol='bar', description='pressure')
    atmosphere = Unit(id=0x26, name='atmosphere', symbol='atm', description='pressure')
    psi = Unit(id=0x27, name='pound force per square inch', symbol='psi', description='pressure')
    becqerel = Unit(id=0x28, name='becqerel', symbol='Bq', description='radioactivity')
    lumen = Unit(id=0x29, name='lumen', symbol='lm', description='light flux')
    lux = Unit(id=0x2A, name='lux', symbol='lx', description='illuminance')
    liter = Unit(id=0x2B, name='liter', symbol='l', description='volume')
    gallon = Unit(id=0x2C, name='gallon (British)', symbol='-', description='volume')
    gallon = Unit(id=0x2D, name='gallon (US liq)', symbol='-', description='volume')
    cubic = Unit(id=0x2E, name='cubic inch', symbol='cu in', description='volume')
    meter_per_sec = Unit(id=0x2F, name='meter per seconds', symbol='m/s', description='speed')
    kmh = Unit(id=0x30, name='kilometre per hour', symbol='km/h', description='speed')
    mph = Unit(id=0x31, name='mile per hour', symbol='mph', description='speed')
    rps = Unit(id=0x32, name='revolutions per second', symbol='rps', description='angular velocity')
    rpm = Unit(id=0x33, name='revolutions per minute', symbol='rpm', description='angular velocity')
    counts = Unit(id=0x34, name='counts', symbol='-', description='-')
    percent = Unit(id=0x35, name='percent', symbol='%', description='-')
    mg_per_stroke = Unit(id=0x36, name='milligram per stroke', symbol='mg/stroke', description='mass per engine stroke')
    meter_per_sec2 = Unit(id=0x37, name='meter per square seconds', symbol='m/s2', description='acceleration')
    Nm = Unit(id=0x38, name='Newton meter', symbol='Nm', description='moment')
    liter_per_min = Unit(id=0x39, name='liter per minute', symbol='l/min', description='flow')
    watt_per_meter2 = Unit(id=0x3A, name='watt per square meter', symbol='W/m2', description='intensity')
    bar_per_sec = Unit(id=0x3B, name='bar per second', symbol='bar/s', description='pressure change')
    radians_per_sec = Unit(id=0x3C, name='radians per second', symbol='rad/s', description='angular velocity')
    radians = Unit(id=0x3D, name='radians square second', symbol='rad/s2', description='angular acceleration')
    kilogram_per_meter2 = Unit(id=0x3E, name='kilogram per square meter', symbol='kg/m2', description='-')
    date1 = Unit(id=0x50, name='Date1', symbol='-', description='Year-Month-Day')
    date2 = Unit(id=0x51, name='Date2', symbol='-', description='Day/Month/Year')
    date3 = Unit(id=0x52, name='Date3', symbol='-', description='Month/Day/Year')
    week = Unit(id=0x53, name='week', symbol='W', description='calendar week')
    time1 = Unit(id=0x54, name='Time1', symbol='-', description='UTC Hour/Minute/Second')
    time2 = Unit(id=0x55, name='Time2', symbol='-', description='Hour/Minute/Second')
    datetime1 = Unit(id=0x56, name='DateAndTime1', symbol='-', description='Second/Minute/Hour/Day/Month/Year')
    datetime2 = Unit(id=0x57, name='DateAndTime2', symbol='-', description='Second/Minute/Hour/Day/Month/Year/Local minute offset/Localhour offset')
    datetime3 = Unit(id=0x58, name='DateAndTime3', symbol='-', description='Second/Minute/Hour/Month/Day/Year')
    datetime4 = Unit(id=0x59, name='DateAndTime4', symbol='-', description='Second/Minute/Hour/Month/Day/Year/Local minute offset/Localhour offset')
